import pandas as pd
import numpy as np

__doc__ = """
Manipulates results to have consistency between the building-scale and district-scale optimizations.
"""


def correct_network_values(reho, scn_id=0, pareto_id=0):
    """
    This function is only useful to find KPIs from the district perspective with a building scale optimization.
    It takes results from the reho object and correct df_KPI, df_Annuals and df_Performance.
    """
    df_grid = reho.results[scn_id][pareto_id]["df_Grid_t"].xs(("Electricity", "Network"), level=("Layer", "Hub"))
    df_export, df_import = get_transformer_import_exports(df_grid)
    nb_periods = reho.cluster["Periods"]

    df = reho.results[scn_id][pareto_id]["df_Grid_t"].sort_index()
    df.loc[("Electricity", "Network"), 'Grid_demand'] = df_export['Grid_profile'].values
    df.loc[("Electricity", "Network"), 'Grid_supply'] = df_import['Grid_profile'].values
    reho.results[scn_id][pareto_id]["df_Grid_t"] = df

    df_grid = reho.results[scn_id][pareto_id]["df_Grid_t"]
    df_time = reho.results[scn_id][pareto_id]["df_Time"]
    surface = reho.ERA
    df_unit_t = reho.results[scn_id][pareto_id]["df_Unit_t"]
    df_KPI = reho.results[scn_id][pareto_id]["df_KPIs"]

    OPEX_m2 = return_correct_OPEX(df_grid, df_time, surface)
    SS_SC = correct_SS_SC(df_grid, df_unit_t, df_time, nb_periods)
    GWP = return_BUI_GWPop(df_grid, surface, df_time, df_KPI)
    GM_GU = correct_grid_param(df_grid)
    results = correct_data_in_reho(reho.results[scn_id][pareto_id], OPEX_m2, SS_SC, GWP, GM_GU, surface)
    return results


def get_transformer_import_exports(df_grid):
    typical_profile = df_grid.Grid_demand - df_grid.Grid_supply  # attention! Supply is negative, feed in is positive
    df = pd.DataFrame(typical_profile, columns=['Grid_profile'])
    df_export = df.copy()
    df_import = df.copy()
    df_export[df_export.Grid_profile < 0] = 0
    df_import[df_import.Grid_profile > 0] = 0
    df_import = -df_import
    return df_export, df_import


def return_annual_exports_imports(df_el, df_t):
    df_aim = {}
    df_export = df_el.Grid_demand.mul(df_t.dp, level='Period', axis=0)
    df_import = df_el.Grid_supply.mul(df_t.dp, level='Period', axis=0)
    df_aim['MWh_imp_el'] = df_import.sum() / 1000  # MWh
    df_aim['MWh_exp'] = df_export.sum() / 1000  # MWh
    return df_aim['MWh_imp_el'], df_aim['MWh_exp']


def get_unit_use_annual_profile(df_unit, df_t, unit, nb_periods):
    df_PV = df_unit[df_unit.index.get_level_values('Unit').str.contains(unit)]
    df_PV = df_PV.groupby('Period').sum()
    df_PV = df_PV.mul(df_t.dp, level='Period', axis=0)
    df_PV = df_PV.loc[0:nb_periods].sum()
    return df_PV


def return_correct_OPEX(df_grid, df_time, total_area):
    opex = 0
    for layer in df_grid.groupby(level=[0]).nunique().index:
        df = df_grid.xs((layer, 'Network'), level=('Layer', 'Hub'))
        df_imp, df_exp = return_annual_exports_imports(df, df_time)
        opex = opex + 1000 * (df.Cost_supply.mean() * df_imp - df.Cost_demand.mean() * df_exp)

    OPEX_m2 = {'opex_m2': opex / total_area}
    return OPEX_m2


def correct_grid_param(df_Grid):
    GM_GU = {}
    df_grid = df_Grid.xs(('Electricity', "Network"), level=('Layer', 'Hub'))[:-2]
    uncontrollable_load = df_Grid.xs("Electricity", level="Layer")["Uncontrollable_load"].drop("Network", level="Hub")
    uncontrollable_load = uncontrollable_load.groupby(level=["Period", "Time"]).sum().max()
    GM_GU['GUd'] = np.round(np.max(df_grid['Grid_demand'] - df_grid['Grid_supply']) / uncontrollable_load, 2)
    GM_GU['GUs'] = np.round(np.max(df_grid['Grid_supply'] - df_grid['Grid_demand']) / uncontrollable_load, 2)

    df = df_Grid.xs('Electricity', level=0)[['Grid_demand', 'Grid_supply']].xs("Network", level="Hub", drop_level=False)
    df_max = df.groupby(level=['Hub', 'Period']).max()
    df_mean = df.groupby(level=['Hub', 'Period']).mean().replace(0, 1)  # replace 0 with 1 to avoid div by 0,  profiles >0, in case av = 0- whole profile is 0
    GM = df_max.div(df_mean, level='Hub').max()

    GM_GU['GMd'] = np.round(GM["Grid_demand"], 2)
    GM_GU['GMs'] = np.round(GM["Grid_supply"], 2)

    return GM_GU


def correct_SS_SC(df_grid, df_unit, df_t, nb_periods=10):
    df_PV = get_unit_use_annual_profile(df_unit, df_t, "PV", nb_periods)
    df_cogen = get_unit_use_annual_profile(df_unit, df_t, "Cogeneration", nb_periods)
    onsite_elec = df_cogen.Units_supply + df_PV.Units_supply

    df_el = df_grid.xs(('Electricity', 'Network'), level=('Layer', 'Hub'))
    df_imp, df_exp = return_annual_exports_imports(df_el, df_t)

    df_aim = {'SC': np.nan if onsite_elec == 0 else (onsite_elec - df_exp * 1000) / onsite_elec,
              'SS': (onsite_elec - df_exp * 1000) / (onsite_elec - df_exp * 1000 + df_imp * 1000),
              "PVP": df_PV.Units_supply / (df_PV.Units_supply - df_exp * 1000 + df_imp * 1000),
              "MWh_imp_el": df_imp,
              "MWh_exp": df_exp}

    return df_aim


def return_BUI_GWPop(df_grid, surface, df_time, df_KPI):
    df_el = df_grid.xs(('Electricity', 'Network'), level=('Layer', 'Hub'))
    emissions_el_dy = df_el.GWP_supply * df_el.Grid_supply - df_el.GWP_demand * df_el.Grid_demand
    emissions_el_dy = emissions_el_dy.mul(df_time.dp, level='Period', axis=0).sum() / surface
    emissions_el_av = df_el.GWP_supply.mean() * df_el.Grid_supply - df_el.GWP_demand.mean() * df_el.Grid_demand
    emissions_el_av = emissions_el_av.mul(df_time.dp, level='Period', axis=0).sum() / surface

    df_resource = df_grid.xs("Network", level="Hub")
    emissions = df_resource.GWP_supply * df_resource.Grid_supply - df_resource.GWP_demand * df_resource.Grid_demand
    emissions = emissions.mul(df_time.dp, level='Period', axis=0).sum()

    gwp_op_m2 = emissions / surface
    gwp_constr_m2 = df_KPI["gwp_constr_m2"].xs("Network")
    gwp_tot_m2 = gwp_op_m2 + gwp_constr_m2

    df_aim = {'gwp_elec_av': emissions_el_av,
              'gwp_elec_dy': emissions_el_dy,
              'gwp_op_m2': gwp_op_m2,
              'gwp_constr_m2': gwp_constr_m2,
              'gwp_tot_m2': gwp_tot_m2}

    return df_aim


def correct_data_in_reho(results, OPEX_m2, SS_SC, GWP, GM_GU, surface):
    df_KPI = results["df_KPIs"]
    df_KPI.at["Network", "SS"] = SS_SC["SS"]
    df_KPI.at["Network", "SC"] = SS_SC["SC"]
    df_KPI.at["Network", "PVP"] = SS_SC["PVP"]
    df_KPI.at["Network", "opex_m2"] = OPEX_m2["opex_m2"]

    df_KPI.at["Network", "gwp_elec_av"] = GWP["gwp_elec_av"]
    df_KPI.at["Network", "gwp_elec_dy"] = GWP["gwp_elec_dy"]
    df_KPI.at["Network", "gwp_op_m2"] = GWP["gwp_op_m2"]
    df_KPI.at["Network", "gwp_constr_m2"] = GWP["gwp_constr_m2"]
    df_KPI.at["Network", "gwp_tot_m2"] = GWP["gwp_tot_m2"]

    df_KPI.at["Network", "GUd"] = GM_GU["GUd"]
    df_KPI.at["Network", "GUs"] = GM_GU["GUs"]
    df_KPI.at["Network", "GMd"] = GM_GU["GMd"]
    df_KPI.at["Network", "GMs"] = GM_GU["GMs"]
    results["df_KPIs"] = df_KPI

    df_perf = results["df_Performance"]
    df_perf.at["Network", "Costs_op"] = OPEX_m2["opex_m2"] * surface
    df_perf.at["Network", "GWP_op"] = GWP["gwp_op_m2"] * surface

    results["df_Annuals"].at[("Electricity", "Network"), "Demand_MWh"] = SS_SC["MWh_exp"]
    results["df_Annuals"].at[("Electricity", "Network"), "Supply_MWh"] = SS_SC["MWh_imp_el"]

    return results



__doc__ = """
REHO - A Comprehensive Decision Support Tool for Sustainable Energy System Planning.
====================================================================================
Renewable Energy Hub Optimizer (REHO) is a decision support tool for sustainable urban energy system planning.
It is developed by EPFL (Switzerland), within the Industrial Process and Energy Systems Engineering (IPESE) group.
"""


import configparser
import csv
import math
import os.path
import re
import warnings

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely import wkt
from sqlalchemy import create_engine, MetaData, select
from sqlalchemy.dialects import postgresql
from sqlalchemy.exc import SAWarning

import reho.model.preprocessing.skydome as skydome
from reho.paths import *

__doc__ = """
Handles data for buildings characterization.
"""


class QBuildingsReader:
    """
    Handles and prepares the data related to buildings.

    There usually come from `GBuildings <https://ipese-web.epfl.ch/lepour/qbuildings/index.html>`_ database. However,
    one can use data from a csv, in which case the column names should correspond to the GBuildings one, described in
    `Processed GBuildings tables <https://ipese-web.epfl.ch/lepour/qbuildings/GBuildings/description.html#processed>`_.

    Parameters
    ----------
    load_facades : bool
        Whether the facades data should be added.
    load_roofs : bool
        Whether the roofs data should be added.
    """

    def __init__(self, load_facades=False, load_roofs=False):

        self.db = None
        self.tables = None
        self.db_schema = None
        self.db_engine = None
        self.connection = None
        self.data = {}
        self.load_facades = load_facades
        self.load_roofs = load_roofs

        self.local_data = dict()
        self.local_data["df_Area"] = pd.read_csv(path_to_areas, header=None)
        self.local_data["df_Cenpts"] = pd.read_csv(path_to_cenpts, header=None)

    def establish_connection(self, db):
        """
        Allows to establish the connection with one of the QBuildings database.

        Parameters
        ----------
        db : str
            Name of the database to which we want to connect

        """
        # Database connection
        file_ini = path_to_qbuildings + "/" + db + ".ini"

        project = configparser.ConfigParser()
        project.read(file_ini)

        # Database
        self.db_schema = "Processed"
        try:
            db_engine_str = 'postgresql+psycopg2://{}:{}@{}:{}/{}'.format(project['database']['username'],
                                                                          project['database']['password'],
                                                                          project['database']['host'],
                                                                          project['database']['port'],
                                                                          project['database']['database'])
            self.db_engine = create_engine(db_engine_str)
            self.connection = self.db_engine.connect()  # test connection
            print('Connected to database')

        except Exception as e:
            print(f'Cannot connect to database engine: {e}')

        if 'database' in project:
            print('\thost: {}\n\tport: {}\n\tdatabase: {}\n\tusername: {}'.format(
                project['database']['host'],
                project['database']['port'],
                project['database']['database'],
                project['database']['username']))

        # input
        warnings.filterwarnings('ignore', category=SAWarning)
        metadata = MetaData(bind=self.db_engine)
        metadata.reflect(schema=self.db_schema)
        self.tables = metadata.tables
        self.db = db

        return

    def read_csv(self, buildings_filename='data/buildings.csv', nb_buildings=None, roofs_filename='data/roofs.csv', facades_filename='data/facades.csv'):
        """
        Reads buildings-related data from CSV files and prepare it for the REHO model.

        If not all the buildings from the file should be extracted, one can give a number of buildings.
        The fields from the files are translated to the corresponding ones used in REHO.

        Parameters
        ----------
        buildings_filename : str
            The filename of the CSV file containing buildings data.
        nb_buildings : int, optional
            The number of buildings to consider. If not provided, all buildings in the file are considered.
        roofs_filename : str, optional
            The filename of the CSV file containing roofs data.
        facades_filename : str, optional
            The filename of the CSV file containing facades data.

        Returns
        -------
        dict
            A dictionary containing the prepared data for the REHO model, including buildings, facades, roofs,
            and shadows if roofs and facades are loaded.

        Notes
        -----
        - If ``nb_buildings`` is not provided, all buildings in the 'buildings' data are considered.
        - If ``load_roofs = True``, `roofs_filename` must be provided, else it is not useful. Same goes for the facades.

        Example
        -------
        >>> from reho.model.reho import *
        >>> reader = QBuildingsReader(load_roofs=True)
        >>> qbuildings_data = reader.read_csv("buildings.csv", roofs_filename="roofs.csv", nb_buildings=7)

        >>> qbuildings_data['buildings_data'].keys()
        dict_keys(['Building1', 'Building2', 'Building3'])

        >>> qbuildings_data['buildings_data']['Building1'].keys()
        dict_keys(['id_class', 'ratio', 'status', 'ERA', 'SolarRoofArea', 'area_facade_m2', 'height_m', 'U_h', 'HeatCapacity', 'T_comfort_min_0', 'Th_supply_0', 'Th_return_0', 'Tc_supply_0', 'Tc_return_0', 'x', 'y', 'z', 'geometry', 'transformer', 'id_building', 'egid', 'period', 'n_p', 'energy_heating_signature_kWh_y', 'energy_cooling_signature_kWh_y', 'energy_hotwater_signature_kWh_y', 'energy_el_kWh_y'])
        """
        self.data['buildings'] = file_reader(buildings_filename)
        self.data['buildings'] = translate_buildings_to_REHO(self.data['buildings'])
        # self.data['buildings'] = add_geometry(self.data['buildings'])

        if nb_buildings is None:
            nb_buildings = self.data['buildings'].shape[0]
        buildings = self.select_buildings_data(nb_buildings, None)
        # buildings = add_geometry(buildings)
        qbuildings = {'buildings_data': buildings}
        if self.load_facades:
            self.data['facades'] = file_reader(path_handler(facades_filename))
            selected_facades = self.select_roofs_or_facades_data(roof=False)
            self.data['facades'] = self.data['facades'][self.data['facades'].index.isin(selected_facades)]
            self.data['facades'] = add_geometry(self.data['facades'])
            self.data['facades'] = translate_facades_to_REHO(self.data['facades'], self.data['buildings'])
            qbuildings['facades_data'] = self.data['facades']
            qbuildings['shadows_data'] = return_shadows_district(qbuildings['buildings_data'], self.data['facades'], self.local_data)

        if self.load_roofs:
            self.data['roofs'] = file_reader(path_handler(roofs_filename))
            selected_roofs = self.select_roofs_or_facades_data(roof=True)
            self.data['roofs'] = self.data['roofs'][self.data['roofs'].index.isin(selected_roofs)]
            self.data['roofs'] = add_geometry(self.data['roofs'])
            self.data['roofs'] = translate_roofs_to_REHO(self.data['roofs'])
            qbuildings['roofs_data'] = self.data['roofs']

        return qbuildings

    def read_db(self, transformer=None, nb_buildings=None, egid=None, to_csv=False):
        """
        Reads the database and extract from it the buildings required, by the LV transformer's ID.

        If not all the buildings from the transformer should be extracted, one can give a number of buildings or if
        the EGIDs are known, pass a list of EGIDs.
        The fields from the database are translated to the corresponding ones used in REHO.

        Parameters
        ----------
        transformer : int
            ID of the transformer on which we want to optimize
        nb_buildings : int
            Number of buildings to select
        egid : list
            To specify a list of buildings to optimize with their EGIDs
        to_csv : bool
            To export the data into csv

        Returns
        -------
        dict
            A dictionary that contains the qbuildings data. The default has only one key ``buildings_data``
            with a dictionary of buildings, with their fields and corresponding values.


        Notes
        -----
        - The use of this function requires the previous creation of a ``QBuildingsReader`` and the use of ``establish_connection('Suisse')``.
        - EGIDs are the postal address unique identifier used in Switzerland. One can find the EGIDs of a given address at the `RegBL <https://www.housing-stat.ch/fr/query/adrtoegid.html>`_.
        - If ``load_roofs = True`` the roofs are returned as well in the dictionary as a DataFrame under the key ``roofs_data``.
        - If ``load_facades = True`` the facades and the shadows are returned as well in the dictionary as a DataFrame under the keys ``roofs_data`` and ``shadows_data``.

        Examples
        --------
        >>> from reho.model.reho import *
        >>> reader = QBuildingsReader(load_roofs=True)
        >>> reader.establish_connection('Suisse')
        >>> qbuildings_data = reader.read_db(transformer=3658, egid=[954117])

        >>> qbuildings_data['buildings_data']
        {'buildings_data': {'Building1': {'id_class': 'I', 'ratio': '1.0', 'status': "['existing', 'existing', 'existing']", 'ERA': 1396.0, 'SolarRoofArea': 1121.8206745917826, 'area_facade_m2': 848.6771960464813, 'height_m': 9.211343577064236, 'U_h': 0.00152, 'HeatCapacity': 120.29999999999991, 'T_comfort_min_0': 20.0, 'Th_supply_0': 65.0, 'Th_return_0': 50.0, 'Tc_supply_0': 12.0, 'Tc_return_0': 17.0, 'x': 2592703.9673297284, 'y': 1120087.7339999992, 'z': 572.4461527539248, 'geometry': <POLYGON ((2592684.383 1120074.623, 2592683.644 1120075.443, 2592679.083 112...>, 'transformer': 3658, 'id_building': '40214', 'egid': '954117', 'period': '1981-1990', 'n_p': 34.9, 'energy_heating_signature_kWh_y': 111855.52745599969, 'energy_cooling_signature_kWh_y': 0.0, 'energy_hotwater_signature_kWh_y': 4562.903646729638, 'energy_el_kWh_y': 39088.0}}

        >>> qbuildings_data['roofs_data']
            TILT  ...                                           geometry
        0     26  ...  MULTIPOLYGON (((2592819.164 1120187.216, 25928...
        1     25  ...  MULTIPOLYGON (((2592832.585 1120154.503, 25928...
        2     25  ...  MULTIPOLYGON (((2592819.164 1120187.216, 25928...
        3     26  ...  MULTIPOLYGON (((2592824.929 1120157.956, 25928...
        0     19  ...  MULTIPOLYGON (((2592378.668 1120324.589, 25923...
        ..   ...  ...                                                ...
        25     0  ...  MULTIPOLYGON (((2592872.699 1120127.178, 25928...
        26     0  ...  MULTIPOLYGON (((2592917.016 1120132.965, 25929...
        27    28  ...  MULTIPOLYGON (((2592891.248 1120129.691, 25928...
        28    26  ...  MULTIPOLYGON (((2592901.604 1120125.591, 25929...
        29    27  ...  MULTIPOLYGON (((2592887.725 1120119.181, 25928...
        [252 rows x 6 columns]

        """

        # TODO: SQL query to select only buildings, roofs and facades of interest

        # Select the right transformer
        sqlQuery = select([self.tables[self.db_schema + '.' + 'transformers']]) \
            .where(self.tables[self.db_schema + '.' + 'transformers'].columns.id == transformer)
        self.data['transformers'] = gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()), con=self.db_engine,
                                                     geom_col='geometry').fillna(np.nan)

        # Select buildings
        sqlQuery = select([self.tables[self.db_schema + '.' + 'buildings']]) \
            .where(self.tables[self.db_schema + '.' + 'buildings'].columns.transformer == transformer)
        self.data['buildings'] = gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()), con=self.db_engine,
                                                  geom_col='geometry').fillna(np.nan)
        mask = (self.data['buildings']['egid'].isnull())
        self.data['buildings'] = self.data['buildings'].loc[~mask, :]

        if nb_buildings is None:
            nb_buildings = self.data['buildings'].shape[0]
        if to_csv:
            self.data['buildings'].to_csv('buildings.csv', index=False)

        self.data['buildings'] = translate_buildings_to_REHO(self.data['buildings'])
        buildings = self.select_buildings_data(nb_buildings, egid)
        if to_csv:
            csv_columns = list(buildings[list(buildings.keys())[0]].keys())
            with open('reho_input.csv', 'w') as csvfile:
                writer = csv.DictWriter(csvfile, csv_columns)
                writer.writeheader()
                for building in buildings:
                    writer.writerow(buildings[building])

        qbuildings = {'buildings_data': buildings}

        if self.load_facades:
            # TODO: Correct the roofs and facades selection with the id filtered by select_buildings_data
            self.data['facades'] = gpd.GeoDataFrame()
            for id in self.data['buildings'].id_building:
                sqlQuery = select([self.tables[self.db_schema + '.' + 'facades']]) \
                    .where(self.tables[self.db_schema + '.' + 'facades'].columns.id_building == id)
                self.data['facades'] = pd.concat(
                    (self.data['facades'], gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()),
                                                            con=self.db_engine, geom_col='geometry').fillna(np.nan)))
            if to_csv:
                self.data['facades'].to_csv('facades.csv', index=False)
            self.data['facades'] = translate_facades_to_REHO(self.data['facades'], self.data['buildings'])
            qbuildings['facades_data'] = self.data['facades']
            qbuildings['shadows_data'] = return_shadows_district(qbuildings["buildings_data"], self.data['facades'], self.local_data)
        if self.load_roofs:
            self.data['roofs'] = gpd.GeoDataFrame()
            for id in self.data['buildings'].id_building:
                sqlQuery = select([self.tables[self.db_schema + '.' + 'roofs']]) \
                    .where(self.tables[self.db_schema + '.' + 'roofs'].columns.id_building == id)
                self.data['roofs'] = pd.concat(
                    (self.data['roofs'], gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()),
                                                          con=self.db_engine, geom_col='geometry').fillna(np.nan)))
            if to_csv:
                self.data['roofs'].to_csv('roofs.csv', index=False)
            self.data['roofs'] = translate_roofs_to_REHO(self.data['roofs'])
            qbuildings['roofs_data'] = self.data['roofs']

        if qbuildings["buildings_data"] == {}:
            raise print("Empty building data")

        return qbuildings

    def select_buildings_data(self, nb_buildings, egid=None):

        if egid is None:
            nb_select = 0
            selected_buildings = []
            reindex = []
            for i, building in self.data['buildings'].iterrows():
                # Only execute optimization for complete dictionary else skip and count
                if re.search('XIII', building['id_class']) is None:
                    selected_buildings.append(building['id_building'])
                    nb_select += 1
                    reindex.append("Building" + str(nb_select))
                if nb_select >= nb_buildings:
                    break
            self.data['buildings'] = self.data['buildings'][
                self.data['buildings']['id_building'].isin(selected_buildings)]
            self.data['buildings'].index = reindex
            if self.db_engine is None:
                self.data['buildings'] = add_geometry(self.data['buildings'])
            buildings_data = self.data['buildings'].to_dict('index')

        else:
            nb_select = 1
            if type(egid) != list:
                egid = [egid]

            buildings_data = gpd.GeoDataFrame()
            for i in egid:
                data_single_bui = self.data['buildings'][self.data['buildings']['egid'] == str(i)]
                data_single_bui.index = ["Building" + str(nb_select)]
                nb_select += 1
                buildings_data = pd.concat([buildings_data, data_single_bui])
            buildings_data = buildings_data.to_dict('index')

        return buildings_data

    def select_roofs_or_facades_data(self, roof):
        selected_data = []
        for i, building in self.data['buildings'].iterrows():
            if roof:
                selected_data += \
                    self.data['roofs'].index[self.data['roofs']['id_building'] == building['id_building']].to_list()
            else:
                selected_data += \
                    self.data['facades'].index[self.data['facades']['id_building'] == building['id_building']].to_list()

        return selected_data


def translate_buildings_to_REHO(df_buildings):

    dict_QBuildings_REHO = {

        #################################################
        # Data strictly necessary for a REHO optimization
        #################################################

        # Data for EUD profiles
        'id_building': 'id_building',
        'egid': 'egid',
        'id_class': 'id_class',
        'class': 'class',
        'ratio': 'ratio',
        'status': 'status',
        'period': 'period',
        'capita_cap': 'n_p',

        # Area
        'area_era_m2': 'ERA',
        'area_roof_solar_m2': 'SolarRoofArea',
        'area_facade_m2': 'area_facade_m2',
        'height_m': 'height_m',  # only for use_facades
        'count_floor': 'count_floor',

        # Heating source
        'source_heating': 'source_heating',
        'source_hotwater': 'source_hotwater',

        # Thermal envelope
        'thermal_transmittance_signature_kW_m2_K': 'U_h',
        'thermal_specific_capacity_Wh_m2_K': 'HeatCapacity',

        # Temperature requirements
        'temperature_interior_C': 'T_comfort_min_0',
        'temperature_heating_supply_C': 'Th_supply_0',
        'temperature_heating_return_C': 'Th_return_0',
        'temperature_cooling_supply_C': 'Tc_supply_0',
        'temperature_cooling_return_C': 'Tc_return_0',

        #############################
        # Data not strictly necessary
        #############################

        # Geographic information
        'x': 'x',
        'y': 'y',
        'z': 'z',
        'geometry': 'geometry',
        'transformer': 'transformer',

        # Annual energy
        'energy_heating_signature_kWh_y': 'energy_heating_signature_kWh_y',
        'energy_cooling_signature_kWh_y': 'energy_cooling_signature_kWh_y',
        'energy_hotwater_signature_kWh_y': 'energy_hotwater_signature_kWh_y',
        'energy_el_kWh_y': 'energy_el_kWh_y',

        # Annual roof and facade irradiance
        'roof_annual_irr_kWh_y': 'roof_annual_irr_kWh_y',
        'facade_annual_irr_kWh_y': 'facade_annual_irr_kWh_y'
    }

    try:
        translated_buildings_data = gpd.GeoDataFrame(geometry=df_buildings['geometry'])
    except TypeError:
        # Convert WKT strings to shapely geometries if needed
        df_buildings['geometry'] = df_buildings['geometry'].apply(lambda x: wkt.loads(x) if isinstance(x, str) else x)
        # Filter out invalid geometries
        df_buildings = df_buildings[df_buildings['geometry'].apply(lambda x: x.is_valid if hasattr(x, 'is_valid') else True)]
        # Create GeoDataFrame
        translated_buildings_data = gpd.GeoDataFrame(geometry=df_buildings['geometry'])

    for key in dict_QBuildings_REHO.keys():
        REHO_index = dict_QBuildings_REHO[key]
        try:
            translated_buildings_data[REHO_index] = df_buildings[key]
        except KeyError:
            print('Key %s not in the dictionary' % key)

    df_buildings = translated_buildings_data

    return df_buildings


def translate_facades_to_REHO(df_facades, df_buildings):

    dict_facades = {'azimuth': 'AZIMUTH',
                    'id_facade': 'Facades_ID',
                    'area_facade_solar_m2': 'AREA',
                    'id_building': 'id_building',
                    # 'cx': 'CX',
                    # 'cy': 'CY',
                    'geometry': 'geometry'}

    try:
        translated_facades_data = gpd.GeoDataFrame(geometry=df_facades['geometry'])
    except TypeError:
        # Convert WKT strings to shapely geometries if needed
        df_facades['geometry'] = df_facades['geometry'].apply(lambda x: wkt.loads(x) if isinstance(x, str) else x)
        # Filter out invalid geometries
        df_facades = df_facades[df_facades['geometry'].apply(lambda x: x.is_valid if hasattr(x, 'is_valid') else True)]
        # Create GeoDataFrame
        translated_facades_data = gpd.GeoDataFrame(geometry=df_facades['geometry'])

    for key in dict_facades.keys():
        REHO_index = dict_facades[key]
        try:
            translated_facades_data[REHO_index] = df_facades[key]
        except KeyError:
            print('Key %s not in the dictionary' % key)
            
    df_facades = translated_facades_data
    df_facades['CX'] = df_facades['geometry'].centroid.x
    df_facades['CY'] = df_facades['geometry'].centroid.y
    df_facades['coord_Z0'] = None

    for i, b in df_buildings.iterrows():
        concerned_facades = df_facades[df_facades['id_building'] == b['id_building']]
        if not pd.isna(b['z']):
            df_facades.loc[concerned_facades.index, 'coord_Z0'] = b['z']

    return df_facades


def translate_roofs_to_REHO(df_roofs):

    dict_roofs = {'tilt': 'TILT',
                  'azimuth': 'AZIMUTH',
                  'id_roof': 'ROOF_ID',
                  'area_roof_solar_m2': 'AREA',
                  'id_building': 'id_building',
                  'geometry': 'geometry'}

    try:
        translated_roofs_data = gpd.GeoDataFrame(geometry=df_roofs['geometry'])
    except TypeError:
        # Convert WKT strings to shapely geometries if needed
        df_roofs['geometry'] = df_roofs['geometry'].apply(lambda x: wkt.loads(x) if isinstance(x, str) else x)
        # Filter out invalid geometries
        df_roofs = df_roofs[df_roofs['geometry'].apply(lambda x: x.is_valid if hasattr(x, 'is_valid') else True)]
        # Create GeoDataFrame
        translated_roofs_data = gpd.GeoDataFrame(geometry=df_roofs['geometry'])

    for key in dict_roofs.keys():
        REHO_index = dict_roofs[key]
        try:
            translated_roofs_data[REHO_index] = df_roofs[key]
        except KeyError:
            print('Key %s not in the dictionary' % key)
        
    df_roofs = translated_roofs_data

    return df_roofs


def get_roofs(self, buildings):
    selected_roofs = []
    for i, building in buildings.iterrows():
        selected_roofs += \
            self.data['roofs'].index[self.data['roofs']['id_building'] == building['id_building']].to_list()
    self.data['roofs'] = self.data['roofs'][self.data['roofs'].index.isin(selected_roofs)]

    return self.data['roofs']


def get_facades(self, buildings):
    selected_facades = []
    for i, building in buildings.iterrows():
        selected_facades += \
            self.data['facades'].index[self.data['facades']['id_building'] == building['id_building']].to_list()
    self.data['facades'] = self.data['facades'][self.data['facades'].index.isin(selected_facades)]

    return self.data['facades']


def calculate_id_building_shadows(df_angles, id_building, local_data):
    df_angles['to_id_building'] = pd.to_numeric(df_angles['to_id_building'])
    df_angles = df_angles.set_index('to_id_building')
    df_angles = df_angles.xs(id_building)

    df_dome = skydome.skydome_to_df(local_data)

    df_shadow = pd.DataFrame()

    for az in df_dome.azimuth.unique():
        df_angles.loc[:, 'cosa2'] = df_angles.apply(lambda x: skydome.f_cos([x['azimuth'], az]), axis=1)
        # df_cosa2 = df_id_building.apply(lambda x: skydome.f_cos([x['azimuth'], az]), axis=1)
        df = df_angles.loc[(df_angles['cosa2'] > 0)].copy()
        # filter buildings which are more than 180 degree apart from patch with az
        df.loc[:, 'tanba'] = df.tanb * df.cosa2
        # calculate tan(beta) for all buildings. Assumption: dxy is the shortest distance and buildings infinite wide

        max_tanba = df['tanba'].max()  # get max obscurance tan(beta, alpha)
        max_b = math.degrees(math.atan(max_tanba))  # get angle

        if math.isnan(max_b):  # if no shadow, set everything to 0
            id_building = 0
            max_b = 0
            max_tanba = 0
        else:
            id_building = df[df.tanba == df['tanba'].max()].index[0]  # get the id_building which causes the obscurance

        df = pd.DataFrame([[max_tanba, max_b, az, id_building]], columns=['tanb', 'beta', 'azimuth', 'id_building'])
        df_shadow = pd.concat((df_shadow, df))

    return df_shadow


def neighbourhood_angles(buildings, facades):
    df_angles = pd.DataFrame()

    for b in buildings:
        id_building = buildings[b]['id_building']
        print('Calculate angles for building: ' + str(id_building))
        df_BUI = pd.DataFrame()
        df_district = {buildings: bui for buildings, bui in buildings.items() if bui['id_building'] != id_building}
        df_district = pd.DataFrame.from_dict(df_district, orient='index')
        df_district = df_district.set_index('id_building')
        # exclude current building to avoid division with zero
        facades_build = facades[facades.id_building == id_building]  # facades of building
        for f in facades_build.index:
            df_c = pd.DataFrame(index=df_district.index)  # df for calculating values for each facade
            df_c['dx'] = df_district.x.values - facades_build.loc[f]['CX']
            df_c['dy'] = df_district.y.values - facades_build.loc[f]['CY']
            df_c['dxy'] = (df_c.dx * df_c.dx + df_c.dy * df_c.dy) ** 0.5
            heights = df_district.height_m.values
            if facades_build.loc[f]['coord_Z0'] is None:
                print('Missing value coord_Z0, not possible to use_facades')
                continue
            df_c['dz'] = df_district.z.values + heights - facades_build.loc[f]['coord_Z0']
            # facades.loc[f]['HEIGHT_Z'] + facades.loc[f]['HEIGHT'] #take foot of facades/ HEIGHT_Z is upperbound
            df_c['tanb'] = df_c.dz / df_c.dxy
            df_c['cosa'] = df_c.dy / df_c.dxy
            df_c['azimuth'] = df_c[['dx', 'dy']].apply(skydome.f_atan, axis=1)
            df_c['beta'] = df_c[['dz', 'dxy']].apply(skydome.f_atan, axis=1)
            df_c = pd.concat([df_c], keys=[f], names=['UID'])
            df_BUI = pd.concat((df_BUI, df_c))

        df_BUI = df_BUI.reset_index()
        df_BUI = df_BUI.rename(columns={'id_building': 'to_id_building'})
        df_BUI['id_building'] = int(id_building)

        df_angles = pd.concat((df_angles, df_BUI))

    df_angles.to_csv('data/angles.csv')

    return df_angles


def return_shadows_district(buildings, facades, local_data):
    df_shadows = pd.DataFrame()

    if os.path.exists('data/angles.csv'):
        df_angles = pd.read_csv('data/angles.csv')
    else:
        df_angles = neighbourhood_angles(buildings, facades)

    for b in buildings:
        id_building = int(buildings[b]['id_building'])
        if id_building in df_angles['id_building'].values:  # check if angle calculation for id_building exists
            df_id_building = calculate_id_building_shadows(df_angles, id_building, local_data)
            idx = np.repeat(id_building, len(df_id_building))
            df_id_building = df_id_building.set_index(idx)
        else:
            print('NO DATA AVAILABLE FOR id_building ' + str(id_building))
            df_id_building = pd.DataFrame(index=[id_building],
                                          columns=['tanb', 'beta', 'azimuth', 'id_building'])  # pass NaN instead
        df_shadows = pd.concat((df_shadows, df_id_building))

    df_shadows["id_building"] = df_shadows["id_building"].astype(str)
    df_shadows.to_csv('data/shadows.csv')

    return df_shadows


def return_shadows_id_building(id_building, df_district, local_data):
    id_building = int(id_building)

    if os.path.isfile('data/shadows.csv'):
        df = file_reader('data/shadows.csv', index_col=0)
    else:
        df = df_district
    df = df.xs(id_building)
    df_dome = skydome.skydome_to_df(local_data)

    df_beta_dome = pd.DataFrame()
    for az in df_dome.azimuth:
        df_beta_dome = pd.concat((df_beta_dome, df.beta[df.azimuth == az]), ignore_index=True)

    df_beta_dome = df_beta_dome.rename(columns={0: 'Limiting_angle_shadow'})

    return df_beta_dome


def add_geometry(df):
    # Avoid issues with geometry when reading data from a csv
    try:
        geom = gpd.GeoSeries.from_wkb(df['geometry'])
    except KeyError:
        print("No geometry in the dataframe")
        return df
    except:
        try:
            geom = gpd.GeoSeries.from_wkt(df['geometry'])
        except TypeError:
            print("Geometry passed is neither of format wkb or wkt so neither from PostGIS, neither from QBuildings")
            return df

    return gpd.GeoDataFrame(df, geometry=geom)



import os
import pandas as pd
import pytest
from reho.test.test_run import test_run
from reho.plotting import plotting


@pytest.fixture(scope="module", autouse=True)
def generate_test_results():
    if not os.path.exists('results/test_results.pickle'):
        test_run(save_results=True)


@pytest.fixture(scope="module")
def results():
    return pd.read_pickle('results/test_results.pickle')


def test_plot_performance(results):
    try:
        plotting.plot_performance(results, plot='costs', indexed_on='Scn_ID', title="Economical performance").show()
        plotting.plot_performance(results, plot='gwp', indexed_on='Scn_ID', title="Environmental performance").show()
        plotting.plot_performance(results, plot='combined', indexed_on='Scn_ID', title="Combined performance (Economical + Environmental)").show()
    except ImportError as e:
        pytest.fail(f"plot_performance failed: {e}")


def test_plot_expenses(results):
    try:
        plotting.plot_expenses(results, plot='costs', indexed_on='Scn_ID', title="Costs and Revenues").show()
    except ImportError as e:
        pytest.fail(f"plot_expenses failed: {e}")


def test_plot_sankey(results):
    try:
        plotting.plot_sankey(results['totex'][0], label='EN_long', color='ColorPastel', title="Sankey diagram").show()
    except ImportError as e:
        pytest.fail(f"plot_sankey failed: {e}")


def test_plot_eud(results):
    try:
        plotting.plot_eud(results, label='EN_long', title="End-use demands").show()
    except ImportError as e:
        pytest.fail(f"plot_eud failed: {e}")


def test_plot_profiles(results):
    try:
        units_to_plot = ['ElectricalHeater', 'HeatPump', 'PV', 'NG_Boiler']
        plotting.plot_profiles(results['totex'][0], units_to_plot, label='EN_long', color='ColorPastel', resolution='weekly',
                               title="Energy profiles with a weekly moving average").show()
    except ImportError as e:
        pytest.fail(f"plot_profiles failed: {e}")


def all_plots(results):
    try:
        test_plot_performance(results)
        test_plot_expenses(results)
        test_plot_sankey(results)
        test_plot_eud(results)
        test_plot_profiles(results)
    except ImportError as e:
        pytest.fail(f"Plotting failed: {e}")



from reho.model.preprocessing.QBuildings import *
import reho.model.preprocessing.weather as weather
from reho.model.preprocessing.emissions_parser import annual_to_typical_emissions

__doc__ = """
Handles data specific to the location.
"""


def return_local_data(cluster, qbuildings_data):
    """
    Retrieves the data (weather and carbon emissions) corresponding to the buildings' location.

    Parameters
    ----------
    cluster : dict
        Defines location of the buildings, and clustering attributes for the data reduction process.
    qbuildings_data : dict
        Buildings characterization

    Returns
    -------
    dict
        File_ID (string) to identify the location and clustering attritutes, and the following pandas dataframes:
            - df_Timestamp
            - df_Irradiation
            - df_Area
            - df_Cenpts
            - df_Westfacades_irr
            - df_Emissions
    """

    local_data = dict()

    # Weather
    File_ID = weather.get_cluster_file_ID(cluster)
    local_data['File_ID'] = File_ID

    path_to_timestamp = os.path.join(path_to_clustering, 'timestamp_' + File_ID + '.dat')
    if not os.path.exists(path_to_timestamp):
        weather.generate_weather_data(cluster, qbuildings_data)

    local_data["df_Timestamp"] = pd.read_csv(path_to_timestamp, delimiter='\t', parse_dates=[0])

    # Solar
    local_data["df_Irradiation"] = pd.read_csv(path_to_irradiation, index_col=[0])
    local_data["df_Area"] = pd.read_csv(path_to_areas, header=None)
    local_data["df_Cenpts"] = pd.read_csv(path_to_cenpts, header=None)

    path_to_westfacades_irr = os.path.join(path_to_clustering, 'westfacades_irr_' + File_ID + '.txt')
    # check if irradiation already exists:
    if not os.path.exists(path_to_westfacades_irr):
        local_data["df_Timestamp"].Date = pd.to_datetime(local_data["df_Timestamp"]['Date'], format="%m/%d/%Y/%H")
        frequency_dict = pd.Series(local_data["df_Timestamp"].Frequency.values, index=local_data["df_Timestamp"].Date).to_dict()
        frequency_dict['PeriodDuration'] = {p + 1: cluster['PeriodDuration'] for p in range(cluster['Periods'])}
        df_annual, irr_west = skydome.calc_orientation_profiles(270, 90, 0, local_data, frequency_dict)
        np.savetxt(path_to_westfacades_irr, irr_west)
    local_data["df_Westfacades_irr"] = pd.read_csv(path_to_westfacades_irr, header=None)[0].values

    # Carbon emissions
    local_data["df_Emissions"] = file_reader(path_to_emissions, index_col=[0, 1, 2])
    local_data["df_Emissions_GWP100a"] = annual_to_typical_emissions(cluster, 'CH', "GWP100a", local_data["df_Timestamp"], local_data["df_Emissions"])

    return local_data



from scipy.stats import qmc

from reho.model.reho import *

__doc__ = """
File for constructing and solving the optimization for the actor-based problem formulation.
"""


class ActorsProblem(REHO):
    """
    Performs an actor-based optimization.

    Parameters are inherited from the REHO class.

    See also
    --------
    reho.model.reho.REHO
    
    Notes
    -------
    This class is still under construction.
    """

    def __init__(self, qbuildings_data, units, grids, parameters=None, set_indexed=None, cluster=None, method=None, scenario=None, solver="highs", DW_params=None):

        super().__init__(qbuildings_data, units, grids, parameters, set_indexed, cluster, method, scenario, solver, DW_params)
    
    def execute_actors_problem(self, n_sample=10, bounds=None, actor="Owners"):
        self.method["actors_problem"] = True
        self.method["include_all_solutions"] = True
        self.method['building-scale'] = True
        self.scenario["Objective"] = "TOTEX_bui"
        self.set_indexed["ActorObjective"] = np.array([actor])
        if bounds is not None:
            sampler = qmc.LatinHypercube(d=2)
            sample = sampler.random(n=n_sample)
            l_bound = [bounds[key][0] for key in ["Utility", "Owners"]]
            u_bound = [bounds[key][1] for key in ["Utility", "Owners"]]
            samples = pd.DataFrame(qmc.scale(sample, l_bound, u_bound), columns=['utility_portfolio', 'owner_portfolio'])
            self.samples = samples.round(4)
        else:
            self.samples = pd.DataFrame([[None, None]] * n_sample, columns=['utility_portfolio', 'owner_portfolio'])
    
        Scn_ID = self.scenario['name']
        self.pool = mp.Pool(mp.cpu_count())
        results = {ids: self.pool.apply_async(self.run_actors_optimization, args=(self.samples, ids)) for ids in self.samples.index}
        while len(results[list(self.samples.index)[-1]].get()) != 3:
            time.sleep(1)
        for ids in self.samples.index:
            df_Results, df_Results_MP, solver_attributes = results[ids].get()
            self.add_df_Results_MP(Scn_ID, ids, self.iter, df_Results_MP, solver_attributes)    # store results MP (pool.apply_async don't store it)
            self.add_df_Results(None, Scn_ID, ids, self.scenario)   # process results based on results MP
            self.get_KPIs(Scn_ID, ids)

        self.samples["objective"] = None
        for i in self.results_MP[self.scenario["name"]]:
            if self.results_MP[self.scenario["name"]][i] is not None:
                self.samples.loc[i, "objective"] = self.results_MP[self.scenario["name"]][i][0]["df_District"]["Objective"]["Network"]
    
        self.pool.close()
    
        gc.collect()  # free memory
    
    def run_actors_optimization(self, samples, ids):
        if any([samples[col][0] for col in samples]):  # if samples contain values
            param = samples.iloc[ids]
            self.parameters = {'utility_portfolio_min': param['utility_portfolio'], 'owner_portfolio_min': param['owner_portfolio']}
        scenario, SP_scenario, SP_scenario_init = self.select_SP_obj_decomposition(self.scenario)
        try:
            scn = self.scenario["name"]
            self.MP_iteration(scenario, Scn_ID=scn, binary=True, Pareto_ID=0)
            self.add_df_Results(None, scn, 0, self.scenario)
            return self.results[scn][0], self.results_MP[scn][0][self.iter], self.solver_attributes_MP
        except:
            return None, None
    
    def read_configurations(self, path=None):
        if path is None:
            filename = 'tr_' + str(self.buildings_data["Building1"]["transformer"]) + '_' + str(len(self.buildings_data)) + '.pickle'
            path = os.path.join(path_to_configurations, filename)
        with open(path, 'rb') as reader:
            [self.results_SP, self.feasible_solutions, self.number_SP_solutions] = pickle.load(reader)
    
    def generate_configurations(self, n_sample=5, tariffs_ranges=None, delta_feed_in=0.15):
        if tariffs_ranges is None:
            tariffs_ranges = {'Electricity': {"Cost_supply_cst": [0.15, 0.45]},
                              'NaturalGas': {"Cost_supply_cst": [0.10, 0.30]}}
        l_bounds = []
        u_bounds = []
        names = []
        for layers in tariffs_ranges:
            for param in tariffs_ranges[layers]:
                l_bounds = l_bounds + [tariffs_ranges[layers][param][0]]
                u_bounds = u_bounds + [tariffs_ranges[layers][param][1]]
                names = names + [layers + '_' + param]
    
        sampler = qmc.LatinHypercube(d=len(l_bounds))
        sample = sampler.random(n=n_sample)
        samples = pd.DataFrame(qmc.scale(sample, l_bounds, u_bounds), columns=names)
        samples = samples.round(3)
        samples['Electricity_Cost_demand_cst'] = samples['Electricity_Cost_supply_cst'] - delta_feed_in
    
        tariffs_ranges['Electricity']['Cost_demand_cst'] = [tariffs_ranges['Electricity']['Cost_supply_cst'][0] - delta_feed_in,
                                                            tariffs_ranges['Electricity']['Cost_supply_cst'][1] - delta_feed_in]
    
        self.pool = mp.Pool(mp.cpu_count())
        for s in samples.index:
            for layer in tariffs_ranges:
                for param in tariffs_ranges[layer]:
                    self.infrastructure.grids[layer][param] = samples.loc[s, layer + "_" + param]
    
            scenario, SP_scenario, SP_scenario_init = self.select_SP_obj_decomposition(self.scenario)
            Scn_ID = self.scenario['name']
            init_beta = [10, 5, 2, 1, 0.5, 0.2, 0.1]
    
            for beta in init_beta:  # execute SP for MP initialization
                if self.method['parallel_computation']:
                    results = {h: self.pool.apply_async(self.SP_initiation_execution, args=(SP_scenario_init, Scn_ID, s, h, None, beta)) for h in
                               self.infrastructure.houses}
                    while len(results[list(self.buildings_data.keys())[-1]].get()) != 2:
                        time.sleep(1)
                    for h in self.infrastructure.houses:
                        (df_Results, attr) = results[h].get()
                        self.add_df_Results_SP(Scn_ID, s, self.iter, h, df_Results, attr)
                self.feasible_solutions += 1  # after each 'round' of SP execution the number of feasible solutions increases
        self.pool.close()
    
        if not os.path.exists(path_to_configurations):
            os.makedirs(path_to_configurations)
    
        filename = 'tr_' + str(self.buildings_data["Building1"]["transformer"]) + '_' + str(len(self.buildings_data)) + '.pickle'
        writer = open(os.path.join(path_to_configurations, filename), 'wb')
        pickle.dump([self.results_SP, self.feasible_solutions, self.number_SP_solutions], writer)



import platform

import matplotlib.pyplot as plt
from SALib.analyze import morris as morris_analyze
from SALib.analyze import sobol as sobol_analyze
from SALib.sample import morris as morris_sample
from SALib.sample import sobol as sobol_sample
from qmcpy import Sobol
from scipy.stats import qmc

from reho.model.reho import *

__doc__ = """
Performs a sensitivity analysis on the optimization.
"""


class SensitivityAnalysis:
    """
    Performs a sensitivity analysis (SA): sampling, solving, storing all optimizations results and the sensitivity of each tested parameter.

    Parameters
    ----------
    reho : reho object
        Model of the district, obtained via the REHO class.
    SA_type : str
        Type of SA, choose between 'Morris', 'Sobol', and 'Monte_Carlo'.
    sampling_parameters : int
        Number of trajectories for the sampling of the solution space.
    upscaling_factor : int
        To represent the effective ERA of the typical districts.

    Notes
    -------
    The framework is designed to be performed using TOTEX minimization but can easily be modified: simply change the objective function in the REHO object initialization, and adapt the calculation for ``objective_values`` in ``extract_results()``.
    """

    def __init__(self, reho, SA_type, sampling_parameters=0, upscaling_factor=1):

        self.district_name = reho.scenario["name"]
        self.reho = reho
        self.SA_type = SA_type
        self.upscaling_factor = upscaling_factor
        self.sampling_parameters = sampling_parameters
        self.ID = str(self.district_name) + "_" + str(self.SA_type) + time.strftime("_%d%m%Y@%Hh%M")
        self.parameter = {}
        self.problem = {}
        self.sampling = []
        self.objective_values = []
        self.SA_results = {'num_optimizations': [], 'dict_df_results': [], 'dict_res_ES': []}
        self.sensitivity = []

    def save(self):
        path_to_SA_results = 'results/'
        file_path = os.path.join(path_to_SA_results, self.ID + '.pickle')
        f = open(file_path, 'wb')
        pickle.dump(self, f)
        f.close()

    def get_lists(self):
        unit_list = self.reho.infrastructure.UnitTypes.tolist()
        try:
            for excluded_unit in self.reho.scenario['exclude_units']:
                unit_list.remove(excluded_unit)
        except ValueError:
            pass
        KPI_list = ['OPEX', 'CAPEX', 'TOTEX', 'GWP']
        return unit_list, KPI_list

    def build_SA(self, unit_parameter=['Cost_inv1', 'Cost_inv2'], SA_parameters={}):
        """
        - Generates the list of parameters for the SA, their values and type of variation range
        - Generates the problem of the SA, i.e. define the parameters and theirs bounds
        - Generates the sampling scheme of the SA

        Parameters
        ----------
        unit_parameter [list]: Units parameters wanted in the SA, by default all parameters are included
        SA_parameters [dict]: Parameters with their bounds set by the script

        Returns
        ---------
        parameter : dict
            Parameters
        problem : dict
            Parameters with their bounds
        sampling : array
            Sampling values
        """

        # 1) Generate the list of parameters

        default_units_values = self.reho.infrastructure.Units_Parameters  # Extract default unit values of the district
        units = np.unique([unit.split('_Building')[0] for unit in self.reho.infrastructure.Units_Parameters.index.to_list()]).tolist()
        if "EV_district" in units:
            units.remove("EV_district")

        for item in self.reho.scenario['exclude_units']:  # Removing excluded units
            try:
                units.remove(item)
            except ValueError:
                pass

        if unit_parameter != []:  # Add units parameters
            for unit in units:
                for parameter in unit_parameter:
                    value = default_units_values[default_units_values.index.str.contains(unit)][parameter].iloc[0]
                    name = str(unit) + "___" + str(parameter)
                    SA_parameters[name] = np.array([0.5, 2]) * value

        self.parameter = SA_parameters
        # 2) Generate a dictionary with all parameters and their bounds for the sampling

        problem = {
            'names': list(SA_parameters.keys()),
            'num_vars': len(SA_parameters.keys()),
            'bounds': list(SA_parameters.values())
        }
        self.problem = problem

        # 3) Description: Generate the sampling of the solution space

        if self.SA_type == "Sobol":
            sampling = sobol_sample.sample(self.problem, self.sampling_parameters, calc_second_order=False)  # Between 500 and 1000, ideally a power of 2
        elif self.SA_type == "Morris":
            sampling = morris_sample.sample(self.problem, self.sampling_parameters)  # Sampling of the space, recommended p = 4 (level) r = 10 (trajectories)
        elif self.SA_type == "Monte_Carlo":
            dimension = len(self.problem["bounds"])
            sampler = Sobol(dimension)
            sample = sampler.gen_samples(self.sampling_parameters)
            l_bounds_values = [bound[0] for bound in self.problem["bounds"]]
            u_bounds_values = [bound[1] for bound in self.problem["bounds"]]
            sampling = qmc.scale(sample, l_bounds_values, u_bounds_values)
        else:
            sampling = None
        self.sampling = sampling

    def run_SA(self, save_inter=True, save_inter_nb_iter=50, save_time_opt=True, intermediate_start=0):
        """
        Launches all optimizations of the SA and store their results.

        Parameters
        -----------

        save_inter : boolean
            Enable intermediary save
        save_inter_nb_iter : int
            Step at which the intermediary save is done
        save_time_opt : boolean
            Ceates a .txt file and write the time for each optimization
        intermediate_start :int
            Starts the SA from a specific sampling point

        Returns
        ---------
        SA_results : dict
            Contains the number of the optimization and a dictionary regrouping all main results of the optimizations
        objective_values : list
            Values of the objective function for each optimization
        """

        path_to_SA_results = 'results/'
        folder = os.path.join(path_to_SA_results, "computational_results")
        if not os.path.exists(path_to_SA_results):
            os.makedirs(path_to_SA_results)
        if not os.path.exists(folder):
            os.makedirs(folder)

        # Extract all attributes of the reho model
        grids = self.reho.infrastructure.grids
        scenario = self.reho.scenario
        district_units = len(self.reho.infrastructure.UnitsOfDistrict) != 0  # True or False
        units = infrastructure.initialize_units(scenario, grids, district_data=district_units)
        n_houses = len(self.reho.buildings_data)

        # Modify the attributes of the model and run SA
        for j in range(intermediate_start, len(self.sampling)):
            print("Optimization number", str(j + 1) + "/" + str(len(self.sampling)))

            sample = self.sampling[j]
            for s, value in enumerate(sample):
                parameter = list(self.parameter.keys())[s]

                if parameter == 'Elec_retail':
                    grids["Electricity"]["Cost_supply_cst"] = value
                elif parameter == 'Elec_feedin':
                    grids["Electricity"]["Cost_demand_cst"] = value
                elif parameter == 'NG_retail':
                    grids["NaturalGas"]["Cost_supply_cst"] = value
                elif parameter == 'Wood_retail':
                    grids["Wood"]["Cost_supply_cst"] = value
                elif parameter == 'Oil_retail':
                    grids["Oil"]["Cost_supply_cst"] = value

                elif "___" in parameter:
                    for unit_id in range(len(units['building_units'])):
                        if units['building_units'][unit_id]['name'] == parameter.split("___")[0]:
                            units['building_units'][unit_id][parameter.split("___")[1]] = value
                else:
                    if parameter in self.reho.lists_MP["list_parameters_MP"]:
                        self.reho.parameters[parameter] = np.array([value])
                    else:
                        self.reho.parameters[parameter] = np.array([value] * n_houses)

            qbuildings_data = {'buildings_data': self.reho.buildings_data}
            self.reho.infrastructure = infrastructure.Infrastructure(qbuildings_data, units, grids)

            try:
                tic = time.perf_counter()
                self.reho.single_optimization(Pareto_ID=j)  # Optimize the modified model
                toc = time.perf_counter()
                time_spent = toc - tic

                self.reho.initialize_optimization_tracking_attributes()

                if save_time_opt:
                    file_name = os.path.join(folder, str(self.SA_type) + '_time.txt')
                    function = 'w' if not os.path.exists(folder) else 'a+'
                    with open(file_name, function) as f:
                        f.write(str(round(time_spent)) + "\n")

                if save_inter:  # Intermediary save
                    if np.mod(j, save_inter_nb_iter) == 0 or j == (len(self.sampling) - 1):
                        self.save()
                        if platform.system() == 'Windows':
                            os.system('cmd /c "ampl_lic restart"')  # restart ampl license to avoid crashes (used with parallel computing)

            except KeyboardInterrupt:
                return

    def calculate_SA(self):
        """
        Computes the sensitivity indices with the objective values and the problem.
        """
        if self.SA_type == "Sobol":
            sensitivity = sobol_analyze.analyze(self.problem, np.array(self.objective_values), print_to_console=True, calc_second_order=False)
        if self.SA_type == "Morris":
            sensitivity = morris_analyze.analyze(self.problem, self.sampling, np.array(self.objective_values), print_to_console=True)
        self.sensitivity = sensitivity

    def plot_Morris(self, save=False):
        fig, ax = plt.subplots(figsize=(8, 8))
        df_ = pd.DataFrame(self.sensitivity).sort_values('mu_star', ascending=False).copy()
        df_.reset_index(inplace=True, drop=True)
        for j, row in df_.iterrows():
            plt.scatter(row['mu_star'], y=row['sigma'], label=row["names"], c='blue', alpha=0.3)
            if row.name < 5:
                plt.annotate(row["names"], xy=(row["mu_star"], row["sigma"]), textcoords="offset pixels",
                             xytext=(20, 20), arrowprops=dict(arrowstyle="-", connectionstyle="arc3"), fontsize=15)
        plt.plot(ax.get_xlim(), ax.get_xlim(), linestyle='dotted', color='black')
        plt.xlabel('$\\mu^*$ [-]', fontsize=18)
        plt.ylabel('$\\sigma$ [-]', fontsize=18)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.title(self.district_name, fontsize=25)
        plt.axis('square')

        if save:
            path_to_SA_results = 'results/'
            plt.savefig(path_to_SA_results + 'Morris_' + self.district_name + '.png', format='png', dpi=300)
        plt.show()

    def extract_results(self, reho, j):
        unit_list, KPI_list = self.get_lists()

        dict_res = {
            'Annual_Network_Exchange': reho.results[self.SA_type][0].df_Annuals.xs("Network", level=1).loc[["Electricity", "NaturalGas"]][
                ['Demand_MWh', 'Supply_MWh']],
            'Elec_Network_t': reho.results[self.SA_type][0].df_Grid_t.xs("Network", level="Hub").xs("Electricity")[['Grid_demand', 'Grid_supply']],
            'NG_Network_t': reho.results[self.SA_type][0].df_Grid_t.xs("Network", level="Hub").xs("NaturalGas")[['Grid_demand', 'Grid_supply']],
            'df_Unit': reho.results[self.SA_type][0].df_Unit['Units_Mult'],
            'Performance_Network': reho.results[self.SA_type][0].df_Performance.xs("Network")
        }

        self.SA_results['num_optimizations'].append(j)
        self.SA_results['dict_df_results'].append(dict_res)
        self.objective_values.append(reho.results[self.SA_type][0].df_Performance['Costs_inv']['Network'] + reho.results[self.SA_type][0].df_Performance['Costs_op']['Network'] + reho.results[self.SA_type][0].df_Performance['Costs_rep']['Network'])

        df_Grid_t = reho.results[self.SA_type][0].df_Grid_t[['Grid_demand', 'Grid_supply']].groupby(['Layer', 'Hub', 'Period']).sum()
        df_Annuals = reho.results[self.SA_type][0].df_Annuals
        df_Unit_t = reho.results[self.SA_type][0].df_Unit_t.groupby(['Layer', 'Unit', 'Period']).sum()

        dict_res_ES = {'E_sector': df_Annuals.groupby(['Layer']).sum(),
                       'InAndOutDistrict': df_Grid_t.xs("Network", level="Hub"),
                       'E_unit': {},
                       'E_resource': df_Unit_t[['Units_demand', 'Units_supply']].groupby(['Layer', 'Period']).sum(),
                       'E_unit_PV': df_Unit_t.xs("Electricity", level="Layer").query('Unit.str.startswith("PV")').groupby(['Period']).sum()['Units_supply']}

        for unit in unit_list:
            dict_res_ES['E_unit'][unit] = df_Annuals.query('Hub.str.startswith("' + str(unit) + '")')['Supply_MWh'].values.sum()

        self.SA_results['dict_res_ES'].append(dict_res_ES)



import os
import requests
from pathlib import Path

BASE_API_URL = "https://api.github.com/repos/IPESE/REHO/contents/scripts/examples"
EXAMPLES_DIR = Path(__file__).parent.parent.parent / "scripts" / "examples"


def download_file(file_url, local_path):
    response = requests.get(file_url)
    if response.status_code == 200:
        with open(local_path, 'wb') as file_out:
            file_out.write(response.content)
        print(f"Downloaded {file_url} to {local_path}")
    else:
        print(f"Failed to download {file_url}: {response.status_code}")


def process_directory(api_url, local_dir):
    response = requests.get(api_url)
    if response.status_code != 200:
        print(f"Failed to fetch directory listing: {response.status_code}")
        return

    items = response.json()
    for item in items:
        item_name = item['name']
        item_path = os.path.join(local_dir, item_name)
        if item['type'] == 'file':
            raw_url = item['download_url']
            download_file(raw_url, item_path)
        elif item['type'] == 'dir':
            os.makedirs(item_path, exist_ok=True)
            process_directory(item['url'], item_path)


def test_download_examples():
    if not EXAMPLES_DIR.exists():
        os.makedirs(EXAMPLES_DIR)
        process_directory(BASE_API_URL, EXAMPLES_DIR)


def execute_script(script_path):
    print(f"Executing {script_path}...")
    os.system(f"python {script_path}")


def test_example_0():
    script_path = Path(EXAMPLES_DIR) / "0_Compact_formulation.py"
    execute_script(script_path)


def test_example_1a():
    script_path = Path(EXAMPLES_DIR) / "1a_Building-scale_totex.py"
    execute_script(script_path)


def test_example_1b():
    script_path = Path(EXAMPLES_DIR) / "1b_Building-scale_Pareto.py"
    execute_script(script_path)


def test_example_2a():
    script_path = Path(EXAMPLES_DIR) / "2a_District-scale_totex.py"
    execute_script(script_path)


def test_example_2b():
    script_path = Path(EXAMPLES_DIR) / "2b_District-scale_Pareto.py"
    execute_script(script_path)


def test_example_3a():
    script_path = Path(EXAMPLES_DIR) / "3a_Read_csv.py"
    execute_script(script_path)


def test_example_3b():
    script_path = Path(EXAMPLES_DIR) / "3b_Custom_infrastructure.py"
    execute_script(script_path)


def test_example_3c():
    script_path = Path(EXAMPLES_DIR) / "3c_HP_T_source.py"
    execute_script(script_path)


def test_example_3d():
    script_path = Path(EXAMPLES_DIR) / "3d_EVs.py"
    execute_script(script_path)


def test_example_3e():
    script_path = Path(EXAMPLES_DIR) / "3e_DHN.py"
    execute_script(script_path)


def test_example_3f():
    script_path = Path(EXAMPLES_DIR) / "3f_Custom_profiles.py"
    execute_script(script_path)


def test_example_3g():
    script_path = Path(EXAMPLES_DIR) / "3g_Stochastic_profiles.py"
    execute_script(script_path)


def test_example_3h():
    script_path = Path(EXAMPLES_DIR) / "3h_Fix_units.py"
    execute_script(script_path)


def test_example_3i():
    script_path = Path(EXAMPLES_DIR) / "3i_Electricity_prices.py"
    execute_script(script_path)


def test_example_4a():
    script_path = Path(EXAMPLES_DIR) / "4a_Progressive_scenarios.py"
    execute_script(script_path)


def test_example_4b():
    script_path = Path(EXAMPLES_DIR) / "4b_Sensitivity_analysis.py"
    execute_script(script_path)


def test_example_5a():
    script_path = Path(EXAMPLES_DIR) / "5a_PV_orientation.py"
    execute_script(script_path)


def test_example_5b():
    script_path = Path(EXAMPLES_DIR) / "5b_PV_facades.py"
    execute_script(script_path)


def test_example_6a():
    script_path = Path(EXAMPLES_DIR) / "6a_Actors_problem.py"
    execute_script(script_path)



import os
from csv import Sniffer
from pathlib import Path
from pandas import read_csv, read_table, read_excel
import sys
from dotenv import load_dotenv

__doc__ = """
File for managing file paths and configurations.
"""


load_dotenv()
if "AMPL_PATH" not in os.environ:
    print("AMPL_PATH is not defined. Please include a .env file at the project root (e.g., AMPL_PATH='C:/AMPL')")

path_to_reho = os.path.dirname(__file__)
path_to_data = os.path.join(path_to_reho, 'data')
path_to_model = os.path.join(path_to_reho, 'model')
path_to_plotting = os.path.join(path_to_reho, 'plotting')

# AMPL model
path_to_ampl_model = os.path.join(path_to_model, 'ampl_model')
path_to_units = os.path.join(path_to_ampl_model, 'units')
path_to_district_units = os.path.join(path_to_ampl_model, 'units', 'district_units')
path_to_units_storage = os.path.join(path_to_ampl_model, 'units', 'storage')
path_to_units_h2 = os.path.join(path_to_ampl_model, 'units', 'h2_units')


# data
path_to_elcom = os.path.join(path_to_data, 'elcom')

path_to_emissions = os.path.join(path_to_data, 'emissions', 'electricity_matrix_2019_reduced.csv')

path_to_infrastructure = os.path.join(path_to_data, 'infrastructure')

path_to_qbuildings = os.path.join(path_to_data, 'QBuildings')

path_to_sia = os.path.join(path_to_data, 'SIA')
path_to_sia_equivalence = os.path.join(path_to_sia, 'sia2024_rooms_sia380_1.csv')
path_to_sia_norms = os.path.join(path_to_sia, 'sia2024_data.xlsx')

path_to_skydome = os.path.join(path_to_data, 'skydome')
path_to_irradiation = os.path.join(path_to_skydome, 'total_irradiation.csv')
path_to_areas = os.path.join(path_to_skydome, 'skyPatchesAreas.txt')  # area of patches
path_to_cenpts = os.path.join(path_to_skydome, 'skyPatchesCenPts.txt')  # location of centre points

path_to_weather = os.path.join(path_to_data, 'weather')

# scripts specific paths
path_to_clustering = os.path.join(os.getcwd(), 'data', 'clustering')
path_to_configurations = os.path.join(os.getcwd(), 'results', 'configurations')


def path_handler(path_given):
    """To handle the path to csv file, absolute path or not"""

    if os.path.isabs(path_given):
        if os.path.isfile(path_given):
            return path_given
        else:
            raise FileNotFoundError('The absolute path that was given is not a valid file.')
    else:
        if os.path.isfile(os.path.realpath(path_given)):
            return os.path.realpath(path_given)
        else:
            raise FileNotFoundError('The relative path that was given is not a valid file.')


def file_reader(file, index_col=None):
    """To read data files correctly, whether there are csv, txt, dat or excel"""
    file = Path(path_handler(file))
    try:
        if file.suffix in ['.csv', '.dat', '.txt']:
            sniffer = Sniffer()
            with open(file, 'r') as f:
                line = next(f).strip()
                delim = sniffer.sniff(line)
            return read_csv(file, sep=delim.delimiter, index_col=index_col)
        elif file.suffix == '.xlsx':
            return read_excel(file)
        else:
            return read_table(file)
    except:
        print('It seems there is a problem when reading the file...\n')
        print("%s" % sys.exc_info()[1])



import calendar

import numpy as np
import pandas as pd

from reho.paths import *

__doc__ = """
Utilities for plotting functions.
"""


# Definition of colors and labels
cm = dict({'ardoise': '#413D3A', 'perle': '#CAC7C7', 'rouge': '#FF0000', 'groseille': '#B51F1F',
           'canard': '#007480', 'leman': '#00A79F', 'salmon': '#FEA993', 'green': '#69B58A', 'yellow': '#FFB100',
           'darkblue': '#1246B5', 'lightblue': '#8bacf4', 'perle_light': '#dad8d8',
           'canard_light': '#4d9ea6', 'leman_light': '#4dc2bc', 'salmon_light': '#fec3b4',
           'groseille_light': "#ff6666", 'darkyellow': '#A57300', 'dark': '#000000',
           'yellow_light': '#FECC93'})

# Colors and labels for units and layers
layout = pd.read_csv(os.path.join(path_to_plotting, 'layout.csv'), index_col='Name').dropna(how='all')


def dict_to_df(results, df):
    t = {(Scn_ID, Pareto_ID): results[Scn_ID][Pareto_ID][df]
         for Scn_ID in results.keys()
         for Pareto_ID in results[Scn_ID].keys()}

    df_merged = pd.concat(t.values(), keys=t.keys(), names=['Scn_ID', 'Pareto_ID'], axis=0)

    return df_merged


def moving_average(data, n):
    return np.convolve(data, np.ones(n), 'valid') / n


def handle_zero_rows(df):
    is_zero_row = (df == 0).all(axis=1)
    return df.loc[~is_zero_row]


def custom_round(value, decimal):
    if decimal == 0:
        rounded_value = int(round(value))
    elif decimal == 1:
        rounded_value = round(value, 1)
    else:
        raise ValueError("decimal argument must be 0 or 1")
    return rounded_value


def merge_handles_labels(ax):
    handles = []
    labels = []
    for axis in ax:
        handle, label = axis.get_legend_handles_labels()
        handles = handles + handle
        labels = labels + label

    handles.reverse()
    labels.reverse()
    by_label = dict(zip(labels, handles))
    return by_label


def monthly_average(results, df_to_extract):
    np_to_extract = np.array([])
    np_month = np.array([])
    ranges = divide_hours_into_months()
    for i in range(1, 366):
        hour = i * 24
        month = len(np_to_extract)
        id_period = results['df_Index'].PeriodOfYear[hour]
        data_id = df_to_extract.xs(id_period)
        np_month = np.concatenate((np_month, data_id))
        if ranges[month][1] == hour:
            np_to_extract = np.append(np_to_extract, np.sum(np_month) / (ranges[month][1] - ranges[month][0]))
            np_month = np.array([])

    return np_to_extract


def divide_hours_into_months():
    num_months = 12

    month_ranges = []
    start_hour = 1
    for month in range(num_months):
        days = calendar.monthrange(2023, month + 1)
        end_hour = start_hour + days[1] * 24 - 1
        month_ranges.append((start_hour, end_hour))
        start_hour = end_hour + 1

    return month_ranges


def prepare_dfs(df_Economics, indexed_on='Scn_ID', neg=False, premium_version=None, additional_data=None, scaling_factor=1):
    """
    This function prepares the dataframes that will be needed for the plot_performance and plot_expenses
    """
    if additional_data is None:
        additional_data = {}
    df_Economics = df_Economics.xs('Network', level='Hub', axis=0)
    df_Economics = df_Economics.groupby(level=indexed_on, sort=False).sum() * scaling_factor
    indexes = df_Economics.index.tolist()

    data_capacities = df_Economics.xs('investment', level='Category', axis=1).transpose()
    data_capacities.index.names = ['Unit']

    if 'isolation' in additional_data:
        data_capacities.loc['Isolation', :] = additional_data['isolation']

    data_capacities = data_capacities.reset_index().merge(layout, left_on="Unit", right_on='Name').set_index("Unit").fillna(0)

    data_resources = df_Economics.xs('operation', level='Category', axis=1).transpose()
    indices = data_resources.index.get_level_values(0)
    new_indices = []
    [new_indices.append(tuple(idx.split("_", 1))) for idx in indices]
    for i, tup in enumerate(new_indices):
        if tup == ('costs', 'Electricity'):
            new_indices[i] = ('costs', 'Electrical_grid')
        elif tup == ('revenues', 'Electricity'):
            new_indices[i] = ('revenues', 'Electrical_grid_feed_in')
    data_resources.index = pd.MultiIndex.from_tuples(new_indices, names=['type', 'Layer'])

    if premium_version is not None:
        data_resources.loc[('avoided', 'solar_premium'), :] = data_resources.loc[('avoided', 'PV_SC')] * (premium_version[0] - premium_version[1]) / \
                                                              premium_version[0]
        data_resources.loc[('revenues', 'solar_value'), :] = data_resources.loc[('revenues', 'Electrical_grid_feed_in')] + data_resources.loc[
            ('avoided', 'PV_SC')] - data_resources.loc[('avoided', 'solar_premium')]
        data_resources = data_resources.drop("PV_SC", level='Layer')
        data_resources = data_resources.drop("Electrical_grid_feed_in", level='Layer')

    data_resources = data_resources.drop("PV", level='Layer')

    if 'mobility' in additional_data:
        data_resources.loc[('costs', 'Gasoline'), :] = additional_data['mobility']

    if 'ict' in additional_data:
        data_resources.loc[('costs', 'Data'), :] = additional_data['ict']
        data_resources.loc[('revenues', 'Data'), :] = 0

    if neg:
        indices = data_resources.index.get_level_values(0)
        neg_indices = indices.str.contains('avoided')
        neg_indices = neg_indices + indices.str.contains('revenues')
        data_resources.loc[neg_indices] = - data_resources.loc[neg_indices]
    data_resources = data_resources.reset_index().merge(layout, left_on='Layer', right_on='Name').set_index(['type', 'Layer'])

    return indexes, data_capacities, data_resources


def remove_building_from_index(df):
    """
    Removes the Building_[123] appended to the name of the units in the index
    """

    def filter_building_str(str):
        str_split = str.split("_")
        if len(str_split) > 2:
            new_idx = "_".join(str.split("_", 2)[:2])
        else:
            new_idx = str_split[0]
        return new_idx

    new_index = []
    index_frame = df.index.to_frame()
    if 'Unit' in index_frame.columns:
        for idx in index_frame['Unit']:
            new_index.append(filter_building_str(idx))
        index_frame['Unit'] = new_index
    if 'Hub' in index_frame.columns:
        for idx in index_frame['Hub']:
            new_index.append(filter_building_str(idx))
        index_frame['Hub'] = new_index

    index_modified = pd.MultiIndex.from_frame(index_frame)

    return df.set_index(index_modified)



# from . import sensitivity_analysis (this line was causing issues with import in example 4b)

__doc__ = """
Directory where the output of the optimization from the AMPL model is extracted and processed to give a ``reho.results`` dictionary.
"""



import pytest


def test_import_reho_modules():
    modules = [
        'reho',
        'reho.model.reho',
        'reho.model.master_problem',
        'reho.model.sub_problem',
        'reho.model.actors_problem',
        'reho.model.infrastructure',
        'reho.model.preprocessing',
        'reho.model.postprocessing',
        'reho.plotting',
    ]

    for module in modules:
        try:
            __import__(module)
        except ImportError as e:
            pytest.fail(f"Importing {module} module failed: {e}")



import pytest
import numpy as np
from reho.model.infrastructure import Infrastructure, initialize_grids, initialize_units


@pytest.fixture(scope="module")
def qbuildings_data():
    return {'buildings_data': {
        'Building1': {'ERA': 192, 'HeatCapacity': 119, 'SolarRoofArea': 140, 'T_comfort_min_0': 20, 'Tc_return_0': 17, 'Tc_supply_0': 12, 'Th_return_0': 50,
                      'Th_supply_0': 65, 'U_h': 0.002, 'area_facade_m2': 144, 'class': 'Residential', 'count_floor': 2, 'egid': '1009515',
                      'energy_cooling_signature_kWh_y': 0, 'energy_el_kWh_y': 4007, 'energy_heating_signature_kWh_y': 20400,
                      'energy_hotwater_signature_kWh_y': 1692, 'facade_annual_irr_kWh_y': 67952, 'geometry': "", 'height_m': 6, 'id_building': '8320',
                      'id_class': 'II', 'n_p': 10, 'period': '1961-1970', 'ratio': '1', 'roof_annual_irr_kWh_y': 154769, 'source_heating': 'Oil',
                      'source_hotwater': 'Oil', 'status': "['existing', 'existing']", 'transformer': 71, 'x': 2496193, 'y': 1114279, 'z': 402},
        'Building2': {'ERA': 117, 'HeatCapacity': 119, 'SolarRoofArea': 101, 'T_comfort_min_0': 20, 'Tc_return_0': 17, 'Tc_supply_0': 12, 'Th_return_0': 50,
                      'Th_supply_0': 65, 'U_h': 0.002, 'area_facade_m2': 121, 'class': 'Residential', 'count_floor': 2, 'egid': '2036614',
                      'energy_cooling_signature_kWh_y': 0, 'energy_el_kWh_y': 2451, 'energy_heating_signature_kWh_y': 12478,
                      'energy_hotwater_signature_kWh_y': 1035, 'facade_annual_irr_kWh_y': 76152, 'geometry': "", 'height_m': 6, 'id_building': '8330',
                      'id_class': 'II', 'n_p': 6, 'period': '1919-1945', 'ratio': '1', 'roof_annual_irr_kWh_y': 118639, 'source_heating': 'Oil',
                      'source_hotwater': 'Electricity', 'status': "['existing', 'existing']", 'transformer': 71, 'x': 2496238, 'y': 1114527, 'z': 405},
    }
    }


@pytest.fixture(scope="module")
def scenario():
    return {'exclude_units': ['NG_Cogeneration'], 'enforce_units': []}


@pytest.fixture(scope="module")
def grids():
    return initialize_grids()


@pytest.fixture(scope="module")
def units(scenario, grids):
    return initialize_units(scenario, grids)


@pytest.fixture(scope="module")
def infrastructure(qbuildings_data, units, grids):
    return Infrastructure(qbuildings_data, units, grids)


def test_infrastructure_not_empty(infrastructure):
    assert not infrastructure.Units_flowrate.empty
    assert not infrastructure.Grids_flowrate.empty
    assert not infrastructure.Grids_Parameters.empty
    assert not infrastructure.Units_Parameters.empty


def test_infrastructure_initialization(infrastructure):
    assert 'HeatCascade' in infrastructure.Layers
    assert 'Electricity' in infrastructure.Layers
    assert 'Building1' in infrastructure.House
    assert 'Building2' in infrastructure.House

    assert set(infrastructure.grids.keys()) == {'Electricity', 'NaturalGas'}
    assert 'NG_Cogeneration' not in infrastructure.UnitTypes
    assert np.array_equal(infrastructure.LayersOfType['HeatCascade'], np.array(['HeatCascade']))
    assert not infrastructure.UnitsOfDistrict


def test_infrastructure_edge_cases(infrastructure):
    with pytest.raises(KeyError):
        Infrastructure({}, {}, {})

    with pytest.raises(KeyError):
        infrastructure.grids['NonExistentGrid']



import numpy as np
import pandas as pd
import math
from datetime import timedelta
from reho.paths import *
import itertools as it
import matplotlib.pyplot as plt

__doc__ = """
Generates a skydome decomposition into patches for PV orientation.
"""


def convert_results_txt_to_csv(load_timesteps):
    """
    Loads one txt file for each hour/timestep in "load_timesteps".
    txt file contains oriented irradiation of each 145 skypatches.
    Combines all hours to one df.
    """

    df = pd.DataFrame()
    for ts in load_timesteps:
        result_file = os.path.join(path_to_skydome, 'results' + str(ts) + '.txt')

        gh_result = np.loadtxt(result_file)
        gh_result = gh_result.reshape(-1, len(gh_result))  # column to row

        df_hour = pd.DataFrame(gh_result, index=[int(ts)])
        df = pd.concat([df, df_hour])

    t1 = pd.to_datetime('1/1/2005', dayfirst=True, infer_datetime_format=True)

    # hour 1 is between 0:00 - 1:00 and is indexed with starting hour so 0:00
    for h in df.index.values:
        df.loc[h, 'time'] = t1 + timedelta(hours=(int(h) - 1))

    df = df.set_index('time')
    output_file = os.path.join(path_to_skydome, 'total_irradiation.csv')
    df.to_csv(output_file)
    print(df)


def skydome_to_df(local_data):
    """
    Reads two txt files: one containing the area and one the position of the center point of the 145 patches,
    which define the skydome. Calculates basic additional values and returns all data in one single df

    The irradiation dome is defined as a composition of patches, defined by their area and their centroids. Those
    patches do not depend on the location. The sky is always decomposed the same way, as done by
    `Lady Bug tool <https://www.ladybug.tools/ladybug/docs/index.html>`_.

    Returns
    -------
       pd.DataFrame
    """
    df_area = local_data["df_Area"]  # area of patches
    df_cenpts = local_data["df_Cenpts"]  # location of centre points

    df_dome = pd.DataFrame()
    df_dome['Area'] = df_area[0]
    df_dome['X'] = df_cenpts[0]
    df_dome['Y'] = df_cenpts[1]
    df_dome['Z'] = df_cenpts[2]

    # Add basic caluclations
    df_dome['XY'] = df_dome[['X', 'Y']].apply(f_sqrt, axis=1)
    df_dome['XYZ'] = df_dome[['XY', 'Z']].apply(f_sqrt, axis=1)
    df_dome['Sin_e'] = round(df_dome['Z'] / df_dome['XYZ'], 4)
    df_dome['Cos_e'] = round(df_dome['XY'] / df_dome['XYZ'], 4)
    df_dome['Cos_a'] = round(df_dome['Y'] / df_dome['XY'], 4)
    df_dome['Sin_a'] = round(df_dome['X'] / df_dome['XY'], 4)
    df_dome['azimuth'] = df_dome[['X', 'Y']].apply(f_atan, axis=1)
    df_dome['elevation'] = df_dome[['Z', 'XY']].apply(f_atan, axis=1)

    return df_dome


def irradiation_to_df(ampl, df_irradiation, df_time):
    """reads Irradiation values of all 145 for the timesteps given in the csv file.
     Converts them to float and returns them as df"""

    # change column name from string to int
    df_irradiation.columns = df_irradiation.columns.astype(int)
    # change values from string to float
    df_irradiation = df_irradiation.infer_objects()

    # parse index as datetime
    df_irradiation.index = pd.to_datetime(df_irradiation.index)

    # get relevant cluster information
    PeriodDuration = ampl.getParameter('TimeEnd').getValues().toPandas()

    # construct Multiindex
    df_p = pd.DataFrame()
    list_timesteps = []

    for p in df_time.index:
        date1 = df_time.xs(p).Date
        end = PeriodDuration.xs(p + 1).TimeEnd  # ampl starts at 1

        date2 = date1 + timedelta(hours=int(end) - 1)
        df_period = df_irradiation.loc[date1:date2]

        for t in np.arange(1, int(end) + 1):  # ampl starts at 1
            list_timesteps.append((p + 1, t))  # create ampl index

        df_p = pd.concat([df_p, df_period])

    idx = pd.MultiIndex.from_tuples(list_timesteps)

    # marry index and data
    df_irradiation = df_p.set_index(idx)

    df = df_irradiation.stack()
    df = df.reorder_levels([2, 0, 1])
    df = pd.DataFrame(df)
    df = df.rename(columns={0: 'Irr'})
    return df


def irradiation_to_df_general(df_irradiation):
    """reads Irradiation values of all 145 for the timesteps given in the csv file.
     Converts them to float and returns them as df"""

    # change column name from string to int
    df_irradiation.columns = df_irradiation.columns.astype(int)
    # change values from string to float
    df_irradiation = df_irradiation.infer_objects()

    return df_irradiation


def irradiation_to_typical_df(typical_days_string, df_profiles):
    """reads Irradiation values of all 145 for the timesteps given in the csv file.
     Converts them to float and returns them as df"""

    df_profiles.set_index('time', inplace=True)
    df_profiles.index = pd.to_datetime(df_profiles.index)

    # select typical days, keep typday index as reference
    df_typical = pd.DataFrame()
    for i, td in enumerate(typical_days_string):
        df_typical = pd.concat([df_typical, df_profiles[td]], sort=True)

    # save profiles in csv
    df_typical.to_csv(os.path.join(path_to_skydome, 'typical_irradiation.csv'))

    return df_typical


def f_sqrt(x):
    """
    Calculates the square root of a vector with two elements
    """
    return math.hypot(x[0], x[1])


def f_atan(x):
    """
    math.atan2(y, x) : Returns atan(y / x), in radians. The result is between -pi and pi. The vector in the plane from
    the origin to point (x, y) makes this angle with the positive X axis. The point of atan2() is that the signs of both
    inputs are known to it, so it can compute the correct quadrant for the angle. For example, atan(1) and atan2(1, 1)
    are both pi/4, but atan2(-1, -1) is -3*pi/4.
    ATTENTION: skydome is clockwise and uses the angle with positive Y axis, so x[0] has to be X and x[1] the Y value
    """
    x[0] = round(x[0], 8)
    x[1] = round(x[1], 8)
    azimuth = math.degrees(math.atan2(x[0], x[1]))

    if azimuth < 0:
        azimuth = azimuth + 360
    else:
        azimuth = azimuth

    return int(round(azimuth))


def f_cos(x):
    a1 = math.radians(x[0])
    a2 = math.radians(x[1])  # cos(-a) = cos(a)
    return math.cos(a1 - a2)


def calc_orientation_profiles(azimuth, tilt, design_lim_angle, local_data, typical_frequency):
    cos_a = round(math.cos(math.radians(azimuth)), 8)
    sin_a = round(math.sin(math.radians(azimuth)), 8)
    sin_y = round(math.sin(math.radians(tilt)), 8)
    cos_y = round(math.cos(math.radians(tilt)), 8)
    print('PANEL ORIENTATION: azimuth ', azimuth, ', tilt ', tilt)

    df_dome = skydome_to_df(local_data)
    df_irradiation = irradiation_to_df_general(local_data["df_Irradiation"])

    df_irradiation_pos = pd.DataFrame()
    df_irradiation_neg = pd.DataFrame()
    for pt in df_irradiation.columns.values:

        azi_pt = df_dome.xs(pt)['azimuth']
        ele_pt = df_dome.xs(pt)['elevation']
        # ------------------------------
        # piecewise linerization skydome
        # ------------------------------
        delta_azi = np.cos(np.radians(azi_pt - azimuth))
        if delta_azi > 0:
            lim_angle = np.rad2deg(np.arctan(delta_azi * np.tan(np.radians(design_lim_angle))))
        else:
            lim_angle = -1

        if (lim_angle > (ele_pt - 6)) & (lim_angle < (ele_pt + 6)):
            linear_factor = 1 - ((lim_angle - (ele_pt - 6)) / 12)

        elif lim_angle >= (ele_pt + 6):
            linear_factor = 0

        elif lim_angle <= (ele_pt - 6):
            linear_factor = 1

        else:
            linear_factor = 'here is a problem linearization of skydome'
        if linear_factor < 0:
            raise 'linear factor negative, changes irradiation direction'

        # calculation orientation in skydome, rotation
        rotation = - sin_a * sin_y * df_dome.xs(pt)['Sin_a'] * df_dome.xs(pt)['Cos_e'] \
                   - sin_y * cos_a * df_dome.xs(pt)['Cos_a'] * df_dome.xs(pt)['Cos_e'] - cos_y * df_dome.xs(pt)['Sin_e']
        irradiation_patch = round(df_irradiation[pt] * rotation * linear_factor, 10)
        if irradiation_patch.min() < 0:
            df_irradiation_neg[pt] = irradiation_patch
        else:
            df_irradiation_pos[pt] = irradiation_patch
    print('Limiting angle design', design_lim_angle)
    print(len(df_irradiation_pos.columns), 'patches can NOT be seen')
    print(len(df_irradiation_neg.columns), 'patches can be seen')

    df_irradiation_panel_t = df_irradiation_neg.sum(axis=1)

    df_irradiation_panel_t.index = pd.to_datetime(df_irradiation_panel_t.index)  # convert index to datetime
    df_irradiation_panel_t = df_irradiation_panel_t.sort_index()

    # construct annual sum
    df_period = np.array([])

    period_duration = typical_frequency.pop('PeriodDuration')

    for number, key in enumerate(list(typical_frequency.keys())[:-2]):
        hours_component = int(period_duration[number + 1])
        end = key + timedelta(hours=hours_component - 1)
        irr_day = -1 * df_irradiation_panel_t.loc[key: end]
        df_period = np.append(df_period, irr_day.values)

    df_period = np.append(df_period, [df_period.min(), df_period.max()])

    return df_irradiation_panel_t, df_period


def calc_orientated_surface(azimuth, tilt, design_lim_angle, local_data, irradiation_file, typical_frequency):

    df_irradiation_panel_t, df_typical = calc_orientation_profiles(azimuth, tilt, design_lim_angle, local_data, irradiation_file, typical_frequency)

    # construct annual sum
    df_period = pd.DataFrame()

    for key, value in typical_frequency.items():
        irr_day = df_irradiation_panel_t.xs(key).sum() * (-value)
        df_period = df_period.append([irr_day])

    annual_irr = round(df_period.sum().values[0] / 1000, 2)

    print('Sum of typical days is', annual_irr, 'kWh/m2')

    return azimuth, tilt, annual_irr


def construct_annual_orientation_df(limiting_angle, local_data):
    azimuth = np.array(range(0, 360))
    tilt = np.array(range(0, 90, 5))

    df = pd.DataFrame()
    for (a, t) in it.product(azimuth, tilt):
        azimuth, tilt, annual_irr = calc_orientated_surface(a, t, limiting_angle, local_data)
        d = {'azimuth': azimuth, 'tilt': tilt, 'irr': annual_irr}
        df = pd.concat([df, d], ignore_index=True)
    print(df)
    filename = 'orientated_irr_linearized' + str(limiting_angle) + '.csv'
    df.to_csv(filename)
    print('Data saved in: ' + filename)


def limiting_angle_for_tilt():
    a = 180
    limit_angle = np.array(range(0, 21, 1))
    tilt_1 = np.array(range(1, 5, 1))
    tilt_2 = np.array(range(5, 95, 5))
    tilt = np.append(tilt_1, tilt_2)

    df = pd.DataFrame()

    for (l, t) in it.product(limit_angle, tilt):
        azimuth, tilt, annual_irr = calc_orientated_surface(a, t, l)

        d = {'tilt': tilt, 'limit_angle': l, 'irr': annual_irr}
        df = pd.concat([df, d], ignore_index=True)
    print(df)
    filename = 'irr_tilt_limiting_angle_azi' + str(a) + '.csv'
    df.to_csv(filename)
    print('Data saved in: ' + filename)


def plot_irr(save_fig):
    cm = plt.cm.get_cmap('Spectral_r')

    # cm = plt.cm.get_cmap('cividis')
    # cm = plt.cm.get_cmap('tab20c')

    # cm = plt.cm.get_cmap('GnBu')

    filename = os.path.join(path_to_skydome, 'orientated_irr.csv')

    df = pd.read_csv(filename, index_col=0)

    df_2 = df[(df['azimuth'] >= 80) & (df['azimuth'] <= 280)]
    df_3 = df_2[(df_2['irr'] >= 950) & (df_2['irr'] <= 1200)]

    condition = np.array(range(0, 90, 5))
    df_3 = df_3.loc[df_3.tilt.isin(condition)]

    #################################################################################################################
    # plotting: for all: plot df, for only tip plot df_3 (next line ax = df_3.plot....) change ax_xlim([])
    ###############################################################################################################

    ax = df.plot.scatter(x='azimuth', y='irr', c='tilt', cmap=cm, alpha=1, edgecolors='none', vmin=0, vmax=90)

    # az_max = df_3.xs(df_3['irr'].idxmax())['azimuth']
    # irr_max = df_3.xs(df_3['irr'].idxmax())['irr']
    # tilt_max = df_3.xs(df_3['irr'].idxmax())['tilt']

    # plt.scatter(az_max, irr_max, color='black')

    # plt.text((az_max+5), irr_max, 'azimuth: '+str(int(az_max))+', ' + 'tilt: '+ str(int(tilt_max)))
    ax.set_xlim([90, 280])
    ax.set_xlim([0, 360])
    ax.set_xlabel('Azimuth angle [degree]')
    f = plt.gcf()

    cax = f.get_axes()[1]
    xax = f.get_axes()[0]

    cax.set_ylabel('Tilt angle [degree]')
    xax.set_ylabel('Annual irradiation density [kWh/m2]')

    plt.tight_layout()

    if save_fig:
        export_format = 'pdf'
        plt.savefig(('all_irr_spectral' + '.' + str(export_format)), format=export_format)
    else:
        plt.show()


if __name__ == '__main__':

    irradiation_file = os.path.join(path_to_skydome, 'typical_irradiation.csv')
    typical_days_string = ['20050921', '20050228', '20050810', '20050313', '20050725',
                           '20050107', '20050911',
                           '20050618']
    typical_frequency = {'20050921': 54, '20050228': 46, '20050810': 17, '20050313': 49, '20050725': 52,
                         '20050107': 68, '20050911': 49, '20050618': 30}

    # thisfile = os.path.join(path_to_clustering, 'timestamp.dat')
    # df = pd.read_csv(thisfile, delimiter='\t')

    # typical_days_string = df.Date.values

    # load_timesteps = np.array(range(1,8760))
    # convert_results_txt_to_csv(load_timesteps)
    # irradiation_to_typical_df(typical_days_string)
    # skydome_to_tilt(tilt = 30)

    # limiting_angle = 10  # = 0 for no irradiation losses
    # construct_annual_orientation_df(limiting_angle)
    # df_good = irradiation_to_df_general(irradiation_csv)
    # df_wrong = irradiation_to_df(irradiation_csv)

    # azimuth = 175
    # tilt = 20
    # design_lim_angle = 20
    # print('design limiting angle:', design_lim_angle)
    calc_orientated_surface(270, 90, 0, irradiation_file, typical_frequency)
    # limiting_angle_for_tilt()
    # df = skydome_to_df()
    # print(df)



import pytest
from reho.model.reho import *
from reho.plotting import plotting


def test_run(save_results=False):

    try:
        # Set building parameters
        reader = QBuildingsReader()
        reader.establish_connection('Geneva')
        qbuildings_data = reader.read_db(71, egid=['1009515'])

        # Select clustering options for weather data
        cluster = {'Location': 'Geneva', 'Attributes': ['T', 'I', 'W'], 'Periods': 10, 'PeriodDuration': 24}

        # Set scenario
        scenario = dict()
        scenario['Objective'] = 'TOTEX'
        scenario['EMOO'] = {}
        scenario['specific'] = []
        scenario['name'] = 'totex'
        scenario['exclude_units'] = ['ThermalSolar', 'NG_Cogeneration']
        scenario['enforce_units'] = []

        # Initialize available units and grids
        grids = infrastructure.initialize_grids()
        units = infrastructure.initialize_units(scenario, grids)

        # Set method options
        method = {'building-scale': True}

        # Run optimization
        reho = REHO(qbuildings_data=qbuildings_data, units=units, grids=grids, cluster=cluster, scenario=scenario, method=method, solver="highs")
        reho.single_optimization()

        if save_results:
            reho.save_results(format=['xlsx', 'pickle'], filename='test_results')

        # Plot results
        plotting.plot_sankey(reho.results['totex'][0], label='EN_long', color='ColorPastel', title="Sankey diagram").show()

    except ImportError as e:
        pytest.fail(f"Running REHO failed: {e}")



import pandas as pd
import numpy as np

__doc__ = """
Collects data from the `SIA Swiss norms <https://www.sia.ch/fr/services/sia-norm/>`_ , which are used to distinguish between eight different building types in their usage and behavior.
"""


def read_sia2024_rooms_sia380_1(digit, df_SIA_380):
    dict_affiliation2digit = {'collective housing': 'I',
                              'individual housing': 'II',
                              'administrative': 'III',
                              'school': 'IV',
                              'commercial': 'V',
                              'restaurant': 'VI',
                              'gathering places': 'VII',
                              'hospital': 'VIII',
                              'industry': 'IX',
                              'shed, warehouse': 'X',
                              'sports facilities': 'XI',
                              'covered swimming-pool': 'XII',
                              'other': 'XIII'}

    df_SIA_380 = df_SIA_380.rename(columns=dict_affiliation2digit)

    return df_SIA_380[digit]


def read_sia_2024_profiles(status, df):
    df_el_appliance = df["calculs"].iloc[:, 1:25]
    df_el_ap_norm = df_el_appliance.div(df_el_appliance.max(axis=1), axis=0)  # normalize profile
    df_el_ap_norm = df_el_ap_norm.fillna(0)

    # status allows to scale the domestic electricity consumption
    if status == 'existing':  # high electricity consumption
        df_existing = df["data"].iloc[:, 38]
        df_el_appliance = df_el_ap_norm.mul(df_existing.values, axis=0)
    elif status == 'standard':  # medium electricity consumption
        df_standard = df["data"].iloc[:, 36]
        df_el_appliance = df_el_ap_norm.mul(df_standard, axis=0)
    elif (status == 'aim') or (status == 'target'):  # low electricity consumption
        df_aim = df["data"].iloc[:, 37]
        df_el_appliance = df_el_ap_norm.mul(df_aim, axis=0)
    else:
        raise 'Building status unknown'

    df_el_light = df["calculs"].iloc[:, 26:50]
    df_el_light.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_el_add = df["calculs"].iloc[:, 52:76]  # additional light profile f.e. for "showrooms"
    df_el_add.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_heat_gain = df["calculs"].iloc[:, 102:126]  # heatgain by people
    df_heat_gain.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_dhw = df["calculs"].iloc[:, 127:151]
    df_dhw.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_occupancy = df["calculs"].iloc[:, 208:232]
    df_occupancy.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    return df_el_add, df_el_light, df_el_appliance, df_dhw, df_occupancy, df_heat_gain


def daily_profiles_with_monthly_deviation(status, rooms, date, df):
    """
    Returns daily profiles for electricity demand, DHW demand, occupancy, electricity heat gains, and heat gains from people.
    The profiles are based on the SIA norms and vary according to the building specifications (rooms, renovation status) and the date (weekday, month).
    """
    # get monthly deviation
    df_months = df['profiles'].iloc[:, 49:61]

    # rename to match python timestamp
    dict_months = {'Janvier': 1, 'Fevrier': 2, 'Mars': 3, 'Avril': 4, 'Mai': 5, 'Juin': 6,
                   'Juillet': 7, 'Aout': 8, 'Septembre': 9, 'Octobre': 10, 'Novembre': 11, 'Decembre': 12}
    df_months = df_months.rename(columns=dict_months)

    monthly_factor = df_months[date.month]

    # get weekly deviation
    df_free = df['profiles'].iloc[:, 61]  # 0 no days unused, 2 = weekend unused, 1 = Sundays unused

    weekly_factor = np.repeat(1, len(rooms))
    if date.weekday() == 6:  # 6 = Sunday
        weekly_factor[(df_free == 1) | (df_free == 2)] = 0  # rooms which are not used on a Sunday
    if date.weekday() == 5:  # 5 = Saturday
        weekly_factor[df_free == 2] = 0  # rooms which are not used on a Saturday

    # get 2024 profiles
    df_el_appliance, df_el_light, df_el_add, df_dhw, df_occupancy, df_heat_gain = read_sia_2024_profiles(status, df)
    df_el = df_el_appliance + df_el_light + df_el_add  # W/m2
    df_el_gain = df_el_appliance + df_el_light
    # adjust for current day of the year
    df_el = df_el.multiply(weekly_factor * monthly_factor, axis=0)
    df_el_gain = df_el_gain.multiply(weekly_factor * monthly_factor, axis=0)
    df_dhw = df_dhw.multiply(weekly_factor * monthly_factor, axis=0)
    df_occupancy = df_occupancy.multiply(weekly_factor * monthly_factor, axis=0)
    df_heat_gain = df_heat_gain.multiply(weekly_factor * monthly_factor, axis=0)

    # get profiles for each room. multiply by usage. nan if not appearing in building
    df_el = df_el.multiply(rooms.values, axis=0).dropna()
    df_el_gain = df_el_gain.multiply(rooms.values, axis=0).dropna()
    df_dhw = df_dhw.multiply(rooms.values, axis=0).dropna()
    df_occupancy = df_occupancy.multiply(rooms.values, axis=0).dropna()
    df_heat_gain = df_heat_gain.multiply(rooms.values, axis=0).dropna()

    # aggregate
    df_profiles = pd.DataFrame()
    df_profiles['electricity_W/m2'] = df_el.sum(axis=0)
    df_profiles['hotwater_l/m2'] = df_dhw.sum(axis=0)
    df_profiles['occupancy'] = df_occupancy.sum(axis=0)
    df_profiles['elecgain_W/m2'] = df_el_gain.sum(axis=0)
    df_profiles['heatgainpeople_W/m2'] = df_heat_gain.sum(axis=0)

    return df_profiles



import os.path

import numpy as np
import pandas as pd

from reho.paths import *

__doc__ = """
File for handling infrastructure parameters.
"""


class Infrastructure:
    """
    Characterizes all the sets and parameters which are connected to buildings, units and grids.

    Parameters
    ----------
    qbuildings_data : dict
        Buildings characterization
    units : dict
        Units characterization
    grids : dict
        Grids characterization
    """

    def __init__(self, qbuildings_data, units, grids):

        self.units = units["building_units"]
        self.houses = {h: {'units': self.units, 'layers': grids} for h in qbuildings_data['buildings_data'].keys()}
        self.grids = grids
        if "district_units" in units:
            self.district_units = units["district_units"]
        else:
            self.district_units = []

        # Sets -------------------------------------------
        self.House = np.array(list(self.houses.keys()))
        self.Units = np.array([])
        self.UnitTypes = np.unique(np.array([self.houses[h]['units'][u]['UnitOfType'] for h in self.houses for u in range(len(self.houses[h]['units']))]))

        UnitTypeDistrict = np.array([self.district_units[i]["UnitOfType"] for i in range(len(self.district_units))])
        self.UnitTypes = np.unique(np.concatenate([self.UnitTypes, UnitTypeDistrict]))
        self.LayerTypes = np.array(['HeatCascade', 'ResourceBalance'])

        self.LayersOfType = {'HeatCascade': np.array(['HeatCascade']),  # default: each building has a heat cascade
                             'ResourceBalance': np.array(list(self.grids.keys()))}
        self.Layers = np.array(list(self.grids.keys()) + ['HeatCascade'])
        self.Services = np.array(['DHW', 'SH', 'Cooling'])

        self.UnitsOfType = {}
        for u in self.UnitTypes:
            self.UnitsOfType[u] = np.array([])

        self.UnitsOfLayer = {}
        for l in self.Layers:
            self.UnitsOfLayer[l] = np.array([])

        self.UnitsOfHouse = {}
        for h in self.House:
            self.UnitsOfHouse[h] = np.array([])

        self.UnitsOfService = {}
        for s in self.Services:
            self.UnitsOfService[s] = np.array([])

        self.UnitsOfDistrict = np.array([])

        self.HousesOfLayer = {}
        for l in self.Layers:
            self.HousesOfLayer[l] = np.array([])

        self.StreamsOfBuilding = {}
        self.StreamsOfUnit = {}
        self.TemperatureSets = {}
        self.lca_kpis = []
        self.Set = {}

        # Parameter --------------------------------
        self.Units_flowrate = pd.DataFrame()
        self.Grids_flowrate = pd.DataFrame()
        self.Grids_Parameters = pd.DataFrame()
        self.Grids_Parameters_lca = pd.DataFrame()
        self.Units_Parameters = pd.DataFrame()
        self.Units_Parameters_lca = pd.DataFrame()
        self.Streams_H = pd.DataFrame()

        self.HP_parameters = {}

        self.generate_structure()
        self.generate_parameter()

    def generate_structure(self):

        # The indexes h_ht, h_mt, h_lt, c_ht state for the discretization of the streams. They are connected to the heat cascade.
        # h_ht: hotstream_hightemperature. h_mt: hotstream_mediumtemperature. h_lt: hotstream_lowtemperature. c_ht: coldstream_hightemperature

        for h in self.House:
            # Units
            for u in self.houses[h]['units']:
                complete_name = u['name'] + '_' + h

                self.Units = np.append(self.Units, [complete_name])
                self.UnitsOfType[u['UnitOfType']] = np.append(self.UnitsOfType[u['UnitOfType']], [complete_name])
                for l in u['UnitOfLayer']:
                    self.UnitsOfLayer[l] = np.append(self.UnitsOfLayer[l], [complete_name])
                self.UnitsOfHouse[h] = np.append(self.UnitsOfHouse[h], [complete_name])
                for s in u['UnitOfService']:
                    self.UnitsOfService[s] = np.append(self.UnitsOfService[s], [complete_name])

                # Streams
                self.StreamsOfBuilding[h] = np.array(
                    [h + '_c_lt', h + '_c_mt', h + '_h_lt'])  # c_mt  c_lt - space heat demand discretized in 2 streams, _- h_lt for cooling
                self.StreamsOfUnit[complete_name] = np.array([])
                for s in u['StreamsOfUnit']:
                    stream = u['name'] + '_' + h + '_' + s
                    self.StreamsOfUnit[complete_name] = np.append(self.StreamsOfUnit[complete_name], stream)

            # Layers
            for l in self.houses[h]['layers']:
                self.HousesOfLayer[l] = np.append(self.HousesOfLayer[l], [h])
                
        # District units
        for u in self.district_units:
            name = u['name']
            self.Units = np.append(self.Units, [name])
            self.UnitsOfDistrict = np.append(self.UnitsOfDistrict, [name])
            self.UnitsOfType[u['UnitOfType']] = np.append(self.UnitsOfType[u['UnitOfType']], [name])
            for l in u['UnitOfLayer']:
                self.UnitsOfLayer[l] = np.append(self.UnitsOfLayer[l], [name])

            for s in u['UnitOfService']:
                self.UnitsOfService[s] = np.append(self.UnitsOfService[s], [name])

            self.StreamsOfUnit[name] = np.array([])
            for s in u['StreamsOfUnit']:
                stream = u['name'] + '_' + s
                self.StreamsOfUnit[name] = np.append(self.StreamsOfUnit[name], stream)

        lca_kpi_list = np.array(file_reader(os.path.join(path_to_infrastructure, "building_units.csv")).columns)
        lca_kpi_list = [key for key in lca_kpi_list if "_1" in key]
        self.lca_kpis = np.array([key.replace("_1", "") for key in lca_kpi_list])
        self.__generate_set_dict()  # generate dictionary containing all sets for AMPL

    def __generate_set_dict(self):

        self.Set['UnitTypes'] = self.UnitTypes
        self.Set['LayerTypes'] = self.LayerTypes
        self.Set['House'] = np.array(list(self.House))
        self.Set['Services'] = self.Services
        self.Set['Units'] = self.Units
        self.Set['UnitsOfType'] = self.UnitsOfType
        self.Set['UnitsOfHouse'] = self.UnitsOfHouse
        self.Set['UnitsOfService'] = self.UnitsOfService
        self.Set['UnitsOfDistrict'] = self.UnitsOfDistrict

        self.Set['Layers'] = self.Layers
        self.Set['LayersOfType'] = self.LayersOfType
        self.Set['UnitsOfLayer'] = self.UnitsOfLayer
        self.Set['HousesOfLayer'] = self.HousesOfLayer
        self.Set['StreamsOfBuilding'] = self.StreamsOfBuilding
        self.Set['StreamsOfUnit'] = self.StreamsOfUnit
        self.Set['Lca_kpi'] = self.lca_kpis

    def generate_parameter(self):
        # Units Flows --
        for h in self.House:

            for u in self.houses[h]['units']:
                df_i = pd.DataFrame()
                df_o = pd.DataFrame()
                complete_name = u['name'] + '_' + h
                for i in u['Units_flowrate_in']:
                    idx = pd.MultiIndex.from_tuples([(i, complete_name)], names=['Layer', 'Unit'])
                    df = pd.DataFrame(u['Units_flowrate_in'][i], index=idx, columns=['Units_flowrate_in'])
                    df_i = pd.concat([df_i, df])
                for o in u['Units_flowrate_out']:
                    idx = pd.MultiIndex.from_tuples([(o, complete_name)], names=['Layer', 'Unit'])
                    df = pd.DataFrame(u['Units_flowrate_out'][o], index=idx, columns=['Units_flowrate_out'])
                    df_o = pd.concat([df_o, df])

                df = pd.concat([df_o, df_i], axis=1)
                self.Units_flowrate = pd.concat([self.Units_flowrate, df])
        for u in self.district_units:
            df_i = pd.DataFrame()
            df_o = pd.DataFrame()
            name = u['name']
            for i in u['Units_flowrate_in']:
                idx = pd.MultiIndex.from_tuples([(i, name)], names=['Layer', 'Unit'])
                df = pd.DataFrame(u['Units_flowrate_in'][i], index=idx, columns=['Units_flowrate_in'])
                df_i = pd.concat([df_i, df])
            for o in u['Units_flowrate_out']:
                idx = pd.MultiIndex.from_tuples([(o, name)], names=['Layer', 'Unit'])
                df = pd.DataFrame(u['Units_flowrate_out'][o], index=idx, columns=['Units_flowrate_out'])
                df_o = pd.concat([df_o, df])

            df = pd.concat([df_o, df_i], axis=1)
            self.Units_flowrate = pd.concat([self.Units_flowrate, df])

        # Units Costs --
        for h in self.House:
            for u in self.houses[h]['units']:
                self.add_unit_parameters(u['name'] + '_' + h, u)

        for u in self.district_units:
            self.add_unit_parameters(u['name'], u)
        self.Units_Parameters_lca.columns = ["lca_kpi_1", "lca_kpi_2"]

        # Grids
        keys = ['Cost_demand_cst', 'Cost_supply_cst', 'GWP_demand_cst', 'GWP_supply_cst', 'Cost_connection']
        lca_impact_demand = [key + "_demand_cst" for key in self.lca_kpis]
        lca_impact_supply = [key + "_supply_cst" for key in self.lca_kpis]
        for g in self.grids:
            for h in self.House:
                idx = pd.MultiIndex.from_tuples([(g, h)], names=['Layer', 'House'])
                df = pd.DataFrame([[self.grids[g]['Grids_flowrate_out'], self.grids[g]['Grids_flowrate_in']]],
                                  index=idx, columns=['Grids_flowrate_out', 'Grids_flowrate_in'])
                self.Grids_flowrate = pd.concat([self.Grids_flowrate, df])

            df = pd.DataFrame([[self.grids[g][key] for key in keys]], index=[g], columns=keys)
            self.Grids_Parameters = pd.concat([self.Grids_Parameters, df])

            df_lca_demand = pd.DataFrame([[float(self.grids[g][key]) for key in lca_impact_demand]], columns=self.lca_kpis).transpose()
            df_lca_supply = pd.DataFrame([[float(self.grids[g][key]) for key in lca_impact_supply]], columns=self.lca_kpis).transpose()
            df_lca = pd.concat([df_lca_demand, df_lca_supply], axis=1)
            df_lca.index.names = ["Lca_kpi"]
            df_lca["ResourceBalances"] = g
            df_lca = df_lca.set_index("ResourceBalances", append=True)
            self.Grids_Parameters_lca = pd.concat([self.Grids_Parameters_lca, df_lca])
        self.Grids_Parameters_lca.columns = ["lca_kpi_demand_cst", "lca_kpi_supply_cst"]

        # HP and AC temperatures
        for h in self.House:

            for u in self.houses[h]['units']:
                if not u['HP_parameters'] in ['nan', 'None', None]:
                    complete_name = u['name'] + '_' + h
                    file = os.path.join(path_to_infrastructure, u['HP_parameters'])
                    if u['UnitOfType'] == 'Air_Conditioner' or u['UnitOfType'] == 'HeatPump':
                        df = pd.read_csv(file, delimiter=';', index_col=[0, 1])
                        df = pd.concat([df], keys=[complete_name])
                        # get index sets of source and sink of HP
                        name, rest = df.columns[0].split('_', 1)
                        self.TemperatureSets[name + '_Tsink'] = np.array(df.index.get_level_values(1).unique())
                        self.TemperatureSets[name + '_Tsource'] = np.array(df.index.get_level_values(2).unique())
                    else:
                        df = pd.read_csv(file, delimiter=';')
                        df.index = [complete_name]

                    if u['HP_parameters'] in self.HP_parameters:

                        self.HP_parameters[u['HP_parameters']] = pd.concat(
                            [self.HP_parameters[u['HP_parameters']], df])
                    else:
                        self.HP_parameters[u['HP_parameters']] = df

        for key in self.TemperatureSets:  # add additional sets from units to total set
            self.Set[key] = self.TemperatureSets[key]

        # TODO select the AC and HP units without number place in the self.units array
        # self.Set['AC_Tsupply'] = np.array(self.units[2]['stream_Tin'])
        # self.Set['HP_Tsupply'] = np.array(self.units[0]['stream_Tin'])

        # Streams

        Hin = {}
        Hout = {}
        for unitstreams in dict(self.StreamsOfUnit, **self.StreamsOfBuilding).values():  # union of dict
            for s in unitstreams:
                if s.count('_h_') == 1:  # check if it's a hot stream
                    Hin[s] = 1
                    Hout[s] = 0
                elif s.count('_c_') == 1:  # check if it's a cold stream
                    Hin[s] = 0
                    Hout[s] = 1
                else:
                    raise ('Stream ' + str(s) + ' cannot be classified as cold or hot')

        dfin = pd.DataFrame.from_dict(Hin, orient='index', columns=['Streams_Hin'])
        dfout = pd.DataFrame.from_dict(Hout, orient='index', columns=['Streams_Hout'])
        self.Streams_H = pd.concat([dfin, dfout], axis=1)

        Streams_set = []
        for h in self.houses:
            for unit in self.UnitsOfHouse[h]:
                Streams_set = np.concatenate([Streams_set, self.StreamsOfUnit[unit]])
            Streams_set = np.concatenate([Streams_set, self.StreamsOfBuilding[h]])
        self.Streams = Streams_set

    def add_unit_parameters(self, complete_name, unit_param):
        keys = ['Units_Fmin', 'Units_Fmax', 'Cost_inv1', 'Cost_inv2', 'lifetime', 'GWP_unit1', 'GWP_unit2']
        lca_impact_1 = [key + "_1" for key in self.lca_kpis]
        lca_impact_2 = [key + "_2" for key in self.lca_kpis]
        df = pd.DataFrame([[unit_param[key] for key in keys]], columns=keys, index=[complete_name])
        self.Units_Parameters = pd.concat([self.Units_Parameters, df])
        df_lca_1 = pd.DataFrame([[unit_param[key] for key in lca_impact_1]], columns=self.lca_kpis).transpose()
        df_lca_2 = pd.DataFrame([[unit_param[key] for key in lca_impact_2]], columns=self.lca_kpis).transpose()
        df_lca = pd.concat([df_lca_1, df_lca_2], axis=1)
        df_lca.index.names = ["Lca_kpi"]
        df_lca["Units"] = complete_name
        df_lca = df_lca.set_index("Units", append=True)
        self.Units_Parameters_lca = pd.concat([self.Units_Parameters_lca, df_lca])


def prepare_units_array(file, exclude_units=[], grids=None):
    """
    Prepares the array that will be used by initialize_units.

    Parameters
    ----------
    file : str
        Name of the file where to find the units' data (building, district or storage).
    exclude_units : list of str
        The units you want to exclude, given through ``initialize_units``.
    grids : dict
        Grids given through ``initialize_units``.

    Returns
    -------
    np.array
        Contains one dictionary in each cell, with the parameters for a specific unit.

    See also
    --------
    initialize_units

    Notes
    -----
    - Make sure the name of the columns you are using are the same as the one from the default files, that can be found
      in ``data/infrastructure``.
    - The name of the units, which will be used as keys, do not matter but the *UnitOfType* must be along a defined
      list of possibilities.
    """

    def transform_into_list(column):
        for idx, row in column.items():
            try:
                new_value = [float(el) for el in row.split('/') if el != '']
            except:
                new_value = [el.strip() for el in row.split('/') if el != '']
            if unit_data.index.get_loc(idx) == 0 and new_value == []:
                unit_data.at[idx, column.name] = ['']
            unit_data.at[idx, column.name] = new_value

    def check_validity(row):
        if len(row['StreamsOfUnit']) > 1 and len(row['stream_Tin']) == 1:
            row['stream_Tin'] = [row['stream_Tin'][0] for el in row['StreamsOfUnit']]
        if len(row['StreamsOfUnit']) > 1 and len(row['stream_Tout']) == 1:
            row['stream_Tout'] = [row['stream_Tout'][0] for el in row['StreamsOfUnit']]
        return row

    unit_data = file_reader(file)
    unit_data = unit_data.set_index("Unit")

    list_of_columns = ['UnitOfLayer', 'UnitOfService', 'StreamsOfUnit', 'Units_flowrate_in', 'Units_flowrate_out',
                       'stream_Tin', 'stream_Tout']
    try:
        unit_data[list_of_columns] = unit_data[list_of_columns].fillna('').astype(str)
        unit_data[list_of_columns].apply(transform_into_list)
    except KeyError:
        raise KeyError('There is a name in the columns of your csv. Make sure the columns correspond to the default'
                       ' files in data/infrastructure.')

    unit_data = unit_data.apply(check_validity, axis=1)
    unit_data['HP_parameters'] = unit_data['HP_parameters'].astype(str)

    units = []
    if grids is None:
        grid_layers = ['Electricity', 'NaturalGas', 'Oil', 'Wood', 'Data', 'Heat', 'HeatCascade']
    else:
        grid_layers = list(grids.keys()) + ['HeatCascade']

    # Some units need to be defined for thermodynamical reasons in the model. They can still be set to 0.
    units_to_keep = ["PV", "WaterTankSH", "WaterTankDHW", "Battery", "ThermalSolar"]

    for idx, row in unit_data.iterrows():
        if all([layer in grid_layers for layer in row["UnitOfLayer"]]):
            if idx not in exclude_units or row['UnitOfType'] in units_to_keep:
                unit_dict = row.to_dict()
                unit_dict['name'] = idx
                flow_in = {}
                flow_out = {}
                for el in row['Units_flowrate_in']:
                    flow_in[el] = 1e6
                    if el not in row['Units_flowrate_out']:
                        flow_out[el] = 0
                for el in row['Units_flowrate_out']:
                    flow_out[el] = 1e6
                    if el not in row['Units_flowrate_in']:
                        flow_in[el] = 0
                unit_dict['Units_flowrate_in'] = flow_in
                unit_dict['Units_flowrate_out'] = flow_out
                units = np.concatenate([units, [unit_dict]])

    return units


def initialize_units(scenario, grids=None, building_data=os.path.join(path_to_infrastructure, "building_units.csv"),
                     district_data=None, storage_data=None):
    """
    Initializes the available units for the energy system.

    Parameters
    ----------
    scenario : dict or None
        A dictionary containing information about the scenario.
    grids : dict or None, optional
        Information about the energy layers considered. If None, ``['Electricity', 'NaturalGas', 'Oil', 'Wood', 'Data', 'Heat']``.
    building_data : str, optional
        Path to the CSV file containing building unit data. Default is 'building_units.csv'.
    district_data : str or bool or None, optional
        Path to the CSV file containing district unit data. If True, district units are initialized with 'district_units.csv'.
        If None, district units will not be considered. Default is None.
    storage_data :  str or bool or None, optional
        Path to the CSV file containing storage unit data. If True, storage units are initialized with 'storage_units.csv'.
        If None, storage units won't be considered. Default is None.

    Returns
    -------
    dict
        Contains building_units and district_units.

    See also
    --------
    initialize_grids

    Notes
    -----
    - The default files are located in ``reho/data/parameters``.
    - The custom files can be given as absolute or relative path.

    Examples
    --------
    >>> units = infrastructure.initialize_units(scenario, grids, building_data="custom_building_units.csv",
    ...                                         district_data="custom_district_units.csv", storage_data=True)
    """

    default_units_to_exclude = ["Air_Conditioner", "DHN_hex", 'HeatPump_Anergy', 'HeatPump_Lake', 'DataHeat_SH']
    if "exclude_units" not in scenario:
        exclude_units = default_units_to_exclude
    else:
        exclude_units = scenario["exclude_units"] + default_units_to_exclude

    building_units = prepare_units_array(building_data, exclude_units, grids)

    # TODO: these storage units are not fully working
    storage_units_to_exclude = ['BESS_IP', 'PTES_S_IP', 'PTES_C_IP', 'H2S_storage', 'H2_compression', 'SOEFC', 'FC']

    exclude_units = exclude_units + storage_units_to_exclude
    if storage_data is True:
        default_storage_units = os.path.join(path_to_infrastructure, "storage_units.csv")
        building_units = np.concatenate([building_units, prepare_units_array(default_storage_units, exclude_units=exclude_units)])
    elif storage_data:
        building_units = np.concatenate([building_units, prepare_units_array(storage_data, exclude_units=exclude_units)])

    if district_data is True:
        district_units = prepare_units_array(os.path.join(path_to_infrastructure, "district_units.csv"), exclude_units, grids)
    elif district_data:
        district_units = prepare_units_array(district_data, exclude_units, grids)
    else:
        district_units = []

    units = {"building_units": building_units, "district_units": district_units}

    return units


def initialize_grids(available_grids={'Electricity': {}, 'NaturalGas': {}},
                     file=os.path.join(path_to_infrastructure, "grids.csv")):
    """
    Initializes grid information for the energy system.

    Parameters
    ----------
    available_grids : dict, optional
        A dictionary specifying the available grids and their parameters. The keys represent grid names,
        and the values are dictionaries containing optional parameters ['Cost_demand_cst',
        'Cost_supply_cst', 'GWP_demand_cst', 'GWP_supply_cst'].
    file : str, optional
        Path to the CSV file containing grid data. Default is 'grids.csv' in the parameters folder.

    Returns
    -------
    dict
        Contains information about the initialized grids.

    See also
    --------
    initialize_units

    Notes
    -----
    - If one wants to use its one custom grid file, he should pay attention that the name of the layer and
      the parameters correspond.
    - Adding a layer in a custom file will not add it to the model as it is not modelized.

    Examples
    --------
    >>> available_grids = {'Electricity': {'Cost_demand_cst': 0.1, 'GWP_supply_cst': 0.05}, 'NaturalGas': {'Cost_supply_cst': 0.15}}
    >>> grids = initialize_grids(available_grids, file="custom_grids.csv")
    """

    grid_data = file_reader(file)
    grid_data = grid_data.set_index("Grid")

    grids = dict()
    for idx, row in grid_data.iterrows():
        if idx in available_grids.keys():
            grid_dict = row.to_dict()
            grid_dict['name'] = idx
            grid_dict['Grids_flowrate_in'] = 1e6 * grid_dict['Grids_flowrate_in']
            grid_dict['Grids_flowrate_out'] = 1e6 * grid_dict['Grids_flowrate_out']
            if 'Cost_demand_cst' in available_grids[idx]:
                grid_dict['Cost_demand_cst'] = available_grids[idx]['Cost_demand_cst']
            if 'Cost_supply_cst' in available_grids[idx]:
                grid_dict['Cost_supply_cst'] = available_grids[idx]['Cost_supply_cst']
            if 'GWP_demand_cst' in available_grids[idx]:
                grid_dict['GWP_demand_cst'] = available_grids[idx]['GWP_demand_cst']
            if 'GWP_supply_cst' in available_grids[idx]:
                grid_dict['GWP_supply_cst'] = available_grids[idx]['GWP_supply_cst']
            if 'Cost_connection' in available_grids[idx]:
                grid_dict['Cost_connection'] = available_grids[idx]['Cost_connection']
            grids[idx] = grid_dict

    return grids



__doc__ = """
Contains plotting functions and code relative to the visualization of REHO results.

- ``layout.csv``: Contains colors and labels to characterize the units and layers of an energy system configuration.
- ``sia380_1.csv``: Contains the translation of building's affectation in roman numbering to labels in the SIA 380/1 norm.
"""
